# sage.doctest: needs sage.rings.number_field
r"""
Lists of Manin symbols over number fields, elements of `\mathbb{P}^1(R/N)`

Lists of elements of `\mathbb{P}^1(R/N)` where `R` is the ring of integers of a number
field `K` and `N` is an integral ideal.

AUTHORS:

- Maite Aranes (2009): Initial version

EXAMPLES:

We define a P1NFList:

::

    sage: x = polygen(QQ, 'x')
    sage: k.<a> = NumberField(x^3 + 11)
    sage: N = k.ideal(5, a^2 - a + 1)
    sage: P = P1NFList(N); P
    The projective line over
     the ring of integers modulo the Fractional ideal (5, a^2 - a + 1)

List operations with the P1NFList:

::

    sage: len(P)
    26
    sage: [p for p in P]
    [M-symbol (0: 1) of level Fractional ideal (5, a^2 - a + 1),
    ...
    M-symbol (1: 2*a^2 + 2*a) of level Fractional ideal (5, a^2 - a + 1)]

The elements of the P1NFList are M-symbols:

::

    sage: type(P[2])
    <class 'sage.modular.modsym.p1list_nf.MSymbol'>

Definition of MSymbols:

::

    sage: alpha = MSymbol(N, 3, a^2); alpha
    M-symbol (3: a^2) of level Fractional ideal (5, a^2 - a + 1)

Find the index of the class of an M-Symbol `(c: d)` in the list:

::

    sage: i = P.index(alpha)
    sage: P[i].c*alpha.d - P[i].d*alpha.c in N
    True

Lift an MSymbol to a matrix in `SL(2, R)`:

::

    sage: alpha = MSymbol(N, a + 2, 3*a^2)
    sage: alpha.lift_to_sl2_Ok()
    [-a - 1, 15*a^2 - 38*a + 86, a + 2, -a^2 + 9*a - 19]
    sage: Ok = k.ring_of_integers()
    sage: M = Matrix(Ok, 2, alpha.lift_to_sl2_Ok())
    sage: det(M)
    1
    sage: M[1][1] - alpha.d in N
    True

Lift an MSymbol from P1NFList to a matrix in `SL(2, R)`

::

    sage: P[3]
    M-symbol (1: -2*a) of level Fractional ideal (5, a^2 - a + 1)
    sage: P.lift_to_sl2_Ok(3)
    [0, -1, 1, -2*a]
"""
# ****************************************************************************
#       Copyright (C) 2009, Maite Aranes <M.T.Aranes@warwick.ac.uk>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.richcmp import richcmp_method, richcmp
from sage.structure.sage_object import SageObject

from sage.misc.search import search

_level_cache = {}  # The info stored here is used in the normalization of MSymbols.


def P1NFList_clear_level_cache():
    """
    Clear the global cache of data for the level ideals.

    EXAMPLES::

        sage: x = polygen(QQ, 'x')
        sage: k.<a> = NumberField(x^3 + 11)
        sage: N = k.ideal(a+1)
        sage: alpha = MSymbol(N, 2*a^2, 5)
        sage: alpha.normalize()
        M-symbol (-4*a^2: 5*a^2) of level Fractional ideal (a + 1)
        sage: sage.modular.modsym.p1list_nf._level_cache
        {Fractional ideal (a + 1): (...)}
        sage: sage.modular.modsym.p1list_nf.P1NFList_clear_level_cache()
        sage: sage.modular.modsym.p1list_nf._level_cache
        {}
    """
    global _level_cache
    _level_cache = {}


@richcmp_method
class MSymbol(SageObject):
    r"""
    The constructor for an M-symbol over a number field.

    INPUT:

    - ``N`` -- integral ideal (the modulus or level)

    - ``c`` -- integral element of the underlying number field or an MSymbol of
      level N

    - ``d`` -- (optional) when present, it must be an integral element such
      that `\langle c\rangle + \langle d\rangle + N = R`, where `R` is the
      corresponding ring of integers

    - ``check`` -- boolean (default: ``True``); if ``check=False`` the
      constructor does not check the condition
      `\langle c\rangle + \langle d\rangle + N = R`

    OUTPUT:

    An M-symbol modulo the given ideal `N`, i.e. an element of the
    projective line `\mathbb{P}^1(R/N)`, where `R` is the ring of integers of
    the underlying number field.

    EXAMPLES::

        sage: x = polygen(QQ, 'x')
        sage: k.<a> = NumberField(x^3 + 11)
        sage: N = k.ideal(a + 1, 2)
        sage: MSymbol(N, 3, a^2 + 1)
        M-symbol (3: a^2 + 1) of level Fractional ideal (2, a + 1)

    We can give a tuple as input:

    ::

        sage: MSymbol(N, (1, 0))
        M-symbol (1: 0) of level Fractional ideal (2, a + 1)

    We get an error if `\langle c\rangle`, `\langle d\rangle` and `N` are not coprime:

    ::

        sage: MSymbol(N, 2*a, a - 1)
        Traceback (most recent call last):
        ...
        ValueError: (2*a, a - 1) is not an element of P1(R/N).
        sage: MSymbol(N, (0, 0))
        Traceback (most recent call last):
        ...
        ValueError: (0, 0) is not an element of P1(R/N).

    Saving and loading works:

    ::

        sage: alpha = MSymbol(N, 3, a^2 + 1)
        sage: loads(dumps(alpha))==alpha
        True
    """
    def __init__(self, N, c, d=None, check=True):
        """
        See ``MSymbol`` for full documentation.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^4 + 13*x - 7)
            sage: N = k.ideal(5)
            sage: MSymbol(N, 0, 6*a)
            M-symbol (0: 6*a) of level Fractional ideal (5)
            sage: MSymbol(N, a^2 + 3, 7)
            M-symbol (a^2 + 3: 7) of level Fractional ideal (5)
        """
        k = N.number_field()
        R = k.ring_of_integers()
        self.__N = N
        if d is None:  # if we give a list (c, d) or an MSymbol as input
            if isinstance(c, MSymbol):
                if c.N() is N:
                    c1 = R(c[0])
                    d1 = R(c[1])
                else:
                    raise ValueError("Cannot change level of an MSymbol")
            else:
                try:
                    c1 = R(c[0])
                    d1 = R(c[1])
                except (ValueError, TypeError):
                    raise TypeError("Unable to create a Manin symbol from %s" % c)
        else:
            try:
                c1 = R(c)
                d1 = R(d)
            except (ValueError, TypeError):
                raise TypeError("Unable to create a Manin symbol from (%s, %s)" % (c, d))
        if check:
            if (c1.is_zero() and d1.is_zero()) or not N.is_coprime(k.ideal(c1, d1)):
                raise ValueError("(%s, %s) is not an element of P1(R/N)." % (c1, d1))
        self.__c, self.__d = (c1, d1)

    def __repr__(self):
        r"""
        Return the string representation of this MSymbol.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: N = k.ideal(3, a - 1)
            sage: MSymbol(N, 3, a)
            M-symbol (3: a) of level Fractional ideal (3, 1/2*a - 1/2)
        """
        return "M-symbol (%s: %s) of level %s" % (self.__c, self.__d, self.__N)

    def _latex_(self):
        r"""
        Return latex representation of ``self``.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^4 + 13*x - 7)
            sage: N = k.ideal(a^3 - 1)
            sage: alpha = MSymbol(N, 3, 5*a^2 - 1)
            sage: latex(alpha) # indirect doctest
            \(3: 5 a^{2} - 1\)
        """
        return r"\(%s: %s\)" % (self.c._latex_(), self.d._latex_())

    def __richcmp__(self, other, op):
        r"""
        Comparison function for objects of the class MSymbol.

        The order is the same as for the underlying lists of lists.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: N = k.ideal(3, a - 1)
            sage: alpha = MSymbol(N, 3, a)
            sage: beta = MSymbol(N, 1, 0)
            sage: alpha < beta
            False
            sage: beta = MSymbol(N, 3, a + 1)
            sage: alpha < beta
            True
        """
        if not isinstance(other, MSymbol):
            raise ValueError("You can only compare with another M-symbol")
        return richcmp([self.__c.list(), self.__d.list()],
                       [other.__c.list(), other.__d.list()], op)

    def N(self):
        r"""
        Return the level or modulus of this MSymbol.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: N = k.ideal(3, a - 1)
            sage: alpha = MSymbol(N, 3, a)
            sage: alpha.N()
            Fractional ideal (3, 1/2*a - 1/2)
        """
        return self.__N

    def tuple(self):
        r"""
        Return the :class:`MSymbol` as a list `(c, d)`.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: N = k.ideal(3, a - 1)
            sage: alpha = MSymbol(N, 3, a); alpha
            M-symbol (3: a) of level Fractional ideal (3, 1/2*a - 1/2)
            sage: alpha.tuple()
            (3, a)
        """
        return self.__c, self.__d

    def __getitem__(self, n):
        r"""
        Indexing function for the list defined by an M-symbol.

        INPUT:

        - ``n`` -- integer (0 or 1, since the list defined by an M-symbol has
          length 2)

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: N = k.ideal(3, a - 1)
            sage: alpha = MSymbol(N, 3, a); alpha
            M-symbol (3: a) of level Fractional ideal (3, 1/2*a - 1/2)
            sage: alpha[0]
            3
            sage: alpha[1]
            a
        """
        return self.tuple()[n]

    def __get_c(self):
        r"""
        Return the first coefficient of the M-symbol.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(a + 1, 2)
            sage: alpha = MSymbol(N, 3, a^2 + 1)
            sage: alpha.c # indirect doctest
            3
        """
        return self.__c
    c = property(__get_c)

    def __get_d(self):
        r"""
        Return the second coefficient of the M-symbol.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(a + 1, 2)
            sage: alpha = MSymbol(N, 3, a^2 + 1)
            sage: alpha.d # indirect doctest
            a^2 + 1
        """
        return self.__d
    d = property(__get_d)

    def lift_to_sl2_Ok(self):
        r"""
        Lift the :class:`MSymbol` to an element of `SL(2, O_k)`, where `O_k` is the ring
        of integers of the corresponding number field.

        OUTPUT:

        A list of integral elements `[a, b, c', d']` that are the entries of
        a `2\times 2` matrix with determinant 1. The lower two entries are congruent
        (modulo the level) to the coefficients `c`, `d` of the :class:`MSymbol` ``self``.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: N = k.ideal(3, a - 1)
            sage: alpha = MSymbol(N, 3*a + 1, a)
            sage: alpha.lift_to_sl2_Ok()
            [0, -1, 1, a]
        """
        return lift_to_sl2_Ok(self.__N, self.__c, self.__d)

    def normalize(self, with_scalar=False):
        r"""
        Return a normalized :class:`MSymbol` (a canonical representative of an element
        of `\mathbb{P}^1(R/N)`) equivalent to ``self``.

        INPUT:

        - ``with_scalar`` -- boolean (default: ``False``)

        OUTPUT:

        - (only if ``with_scalar=True``) a transforming scalar `u`, such that
          `(u*c', u*d')` is congruent to `(c: d)` (mod `N`), where `(c: d)`
          are the coefficients of ``self`` and `N` is the level.

        -  a normalized :class:`MSymbol` `(c': d')` equivalent to ``self``.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: N = k.ideal(3, a - 1)
            sage: alpha1 = MSymbol(N, 3, a); alpha1
            M-symbol (3: a) of level Fractional ideal (3, 1/2*a - 1/2)
            sage: alpha1.normalize()
            M-symbol (0: 1) of level Fractional ideal (3, 1/2*a - 1/2)
            sage: alpha2 = MSymbol(N, 4, a + 1)
            sage: alpha2.normalize()
            M-symbol (1: -a) of level Fractional ideal (3, 1/2*a - 1/2)

        We get the scaling factor by setting ``with_scalar=True``:

        ::

            sage: alpha1.normalize(with_scalar=True)
            (a, M-symbol (0: 1) of level Fractional ideal (3, 1/2*a - 1/2))
            sage: r, beta1 = alpha1.normalize(with_scalar=True)
            sage: r*beta1.c - alpha1.c in N
            True
            sage: r*beta1.d - alpha1.d in N
            True
            sage: r, beta2 = alpha2.normalize(with_scalar=True)
            sage: r*beta2.c - alpha2.c in N
            True
            sage: r*beta2.d - alpha2.d in N
            True
        """
        N = self.__N
        k = N.number_field()
        R = k.ring_of_integers()

        if self.__c in N:
            if with_scalar:
                return N.reduce(self.d), MSymbol(N, 0, 1)
            return MSymbol(N, 0, 1)
        if self.d in N:
            if with_scalar:
                return N.reduce(self.c), MSymbol(N, 1, 0)
            return MSymbol(N, 1, 0)
        if N.is_coprime(self.c):
            cinv = R(self.c).inverse_mod(N)
            if with_scalar:
                return N.reduce(self.c), MSymbol(N, 1, N.reduce(self.d*cinv))
            return MSymbol(N, 1, N.reduce(self.d*cinv))

        if N in _level_cache:
            Lfacs, Lxs = _level_cache[N]
        else:
            Lfacs = [p**e for p, e in N.factor()]
            Lxs = [(N/p).element_1_mod(p) for p in Lfacs]
            # Lfacs, Lxs only depend of the ideal: same lists every time we
            # call normalize for a given level, so we store the lists.
            _level_cache[N] = (Lfacs, Lxs)
        u = 0  # normalizer factor
        p_i = 0
        for p in Lfacs:
            if p.is_coprime(self.c):
                inv = self.c.inverse_mod(p)
            else:
                inv = self.d.inverse_mod(p)
            u = u + inv*Lxs[p_i]
            p_i = p_i + 1
        c, d = (N.reduce(u*self.c), N.reduce(u*self.d))
        if (c - 1) in N:
            c = R(1)
        if with_scalar:
            return u.inverse_mod(N), MSymbol(N, c, d)
        return MSymbol(N, c, d)


# ************************************************************************
#        P1NFList class                                                  *
# ************************************************************************


@richcmp_method
class P1NFList(SageObject):
    r"""
    The class for `\mathbb{P}^1(R/N)`, the projective line modulo `N`, where
    `R` is the ring of integers of a number field `K` and `N` is an integral ideal.

    INPUT:

    - ``N`` -- integral ideal (the modulus or level)

    OUTPUT:

    A :class:`P1NFList` object representing `\mathbb{P}^1(R/N)`.

    EXAMPLES::

        sage: x = polygen(QQ, 'x')
        sage: k.<a> = NumberField(x^3 + 11)
        sage: N = k.ideal(5, a + 1)
        sage: P = P1NFList(N); P
        The projective line over the ring of integers modulo the Fractional ideal (5, a + 1)

    Saving and loading works.

    ::

        sage: loads(dumps(P)) == P
        True
    """
    def __init__(self, N):
        r"""
        The constructor for the class P1NFList. See ``P1NFList`` for full
        documentation.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 5)
            sage: N = k.ideal(3, a - 1)
            sage: P = P1NFList(N); P
            The projective line over the ring of integers modulo the Fractional ideal (3, a + 2)
        """
        self.__N = N
        self.__list = p1NFlist(N)
        self.__list.sort()

    def __richcmp__(self, other, op):
        r"""
        Comparison function for objects of the class P1NFList.

        The order is the same as for the underlying modulus.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: N1 = k.ideal(3, a + 1)
            sage: P1 = P1NFList(N1)
            sage: N2 = k.ideal(a + 2)
            sage: P2 = P1NFList(N2)
            sage: P1 < P2
            True
            sage: P1 > P2
            False
            sage: P1 == P1NFList(N1)
            True
        """
        if not isinstance(other, P1NFList):
            raise ValueError("You can only compare with another P1NFList")
        return richcmp(self.__N, other.__N, op)

    def __getitem__(self, n):
        r"""
        Standard indexing function for the class P1NFList.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(a)
            sage: P = P1NFList(N)
            sage: list(P) == P._P1NFList__list
            True
            sage: j = randint(0,len(P)-1)
            sage: P[j] == P._P1NFList__list[j]
            True
        """
        return self.__list[n]

    def __len__(self):
        r"""
        Return the length of this P1NFList.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(5, a^2 - a + 1)
            sage: P = P1NFList(N)
            sage: len(P)
            26
        """
        return len(self.__list)

    def __repr__(self):
        r"""
        Return the string representation of this P1NFList.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(5, a + 1)
            sage: P = P1NFList(N); P
            The projective line over the ring of integers modulo the Fractional ideal (5, a + 1)
        """
        return "The projective line over the ring of integers modulo the %s" % self.__N

    def list(self):
        r"""
        Return the underlying list of this :class:`P1NFList` object.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(5, a+1)
            sage: P = P1NFList(N)
            sage: type(P)
            <class 'sage.modular.modsym.p1list_nf.P1NFList'>
            sage: type(P.list())
            <... 'list'>
        """
        return self.__list

    def normalize(self, c, d=None, with_scalar=False):
        r"""
        Return a normalised element of `\mathbb{P}^1(R/N)`.

        INPUT:

        - ``c`` -- integral element of the underlying number field, or an
          MSymbol

        - ``d`` -- (optional) when present, it must be an integral element of
          the number field such that `(c, d)` defines an M-symbol of level `N`

        - ``with_scalar`` -- boolean (default: ``False``)

        OUTPUT:

        - (only if ``with_scalar=True``) a transforming scalar `u`, such that
          `(u*c', u*d')` is congruent to `(c: d)` (mod `N`).

        - a normalized :class:`MSymbol` `(c': d')` equivalent to `(c: d)`.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 31)
            sage: N = k.ideal(5, a + 3)
            sage: P = P1NFList(N)
            sage: P.normalize(3, a)
            M-symbol (1: 2*a) of level Fractional ideal (5, 1/2*a + 3/2)

        We can use an :class:`MSymbol` as input:

        ::

            sage: alpha = MSymbol(N, 3, a)
            sage: P.normalize(alpha)
            M-symbol (1: 2*a) of level Fractional ideal (5, 1/2*a + 3/2)

        If we are interested in the normalizing scalar:

        ::

            sage: P.normalize(alpha, with_scalar=True)
            (-a, M-symbol (1: 2*a) of level Fractional ideal (5, 1/2*a + 3/2))
            sage: r, beta = P.normalize(alpha, with_scalar=True)
            sage: (r*beta.c - alpha.c in N) and (r*beta.d - alpha.d in N)
            True
        """
        if d is None:
            try:
                c = MSymbol(self.__N, c)  # check that c is an MSymbol
            except ValueError:  # catch special case of wrong level
                raise ValueError("The MSymbol is of a different level")
            return c.normalize(with_scalar)
        return MSymbol(self.N(), c, d).normalize(with_scalar)

    def N(self):
        r"""
        Return the level or modulus of this :class:`P1NFList`.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 31)
            sage: N = k.ideal(5, a + 3)
            sage: P = P1NFList(N)
            sage: P.N()
            Fractional ideal (5, 1/2*a + 3/2)
        """
        return self.__N

    def index(self, c, d=None, with_scalar=False):
        r"""
        Return the index of the class of the pair `(c, d)` in the fixed list
        of representatives of `\mathbb{P}^1(R/N)`.

        INPUT:

        - ``c`` -- integral element of the corresponding number field, or an
          :class:`MSymbol`

        - ``d`` -- (optional) when present, it must be an integral element of
          the number field such that `(c, d)` defines an M-symbol of level `N`

        - ``with_scalar`` -- boolean (default: ``False``)

        OUTPUT:

        - ``u`` -- the normalizing scalar (only if ``with_scalar=True``)

        - ``i`` -- the index of `(c, d)` in the list

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 31)
            sage: N = k.ideal(5, a + 3)
            sage: P = P1NFList(N)
            sage: P.index(3,a)
            5
            sage: P[5]==MSymbol(N, 3, a).normalize()
            True

        We can give an :class:`MSymbol` as input:

        ::

            sage: alpha = MSymbol(N, 3, a)
            sage: P.index(alpha)
            5

        We cannot look for the class of an :class:`MSymbol` of a different level:

        ::

            sage: M = k.ideal(a + 1)
            sage: beta = MSymbol(M, 0, 1)
            sage: P.index(beta)
            Traceback (most recent call last):
            ...
            ValueError: The MSymbol is of a different level

        If we are interested in the transforming scalar:

        ::

            sage: alpha = MSymbol(N, 3, a)
            sage: P.index(alpha, with_scalar=True)
            (-a, 5)
            sage: u, i = P.index(alpha, with_scalar=True)
            sage: (u*P[i].c - alpha.c in N) and (u*P[i].d - alpha.d in N)
            True
        """
        if d is None:
            try:
                c = MSymbol(self.__N, c)  # check that c is an MSymbol
            except ValueError:  # catch special case of wrong level
                raise ValueError("The MSymbol is of a different level")
            if with_scalar:
                u, norm_c = c.normalize(with_scalar=True)
            else:
                norm_c = c.normalize()
        else:
            if with_scalar:
                u, norm_c = MSymbol(self.__N, c, d).normalize(with_scalar=True)
            else:
                norm_c = MSymbol(self.__N, c, d).normalize()
        t, i = search(self.__list, norm_c)
        if t:
            if with_scalar:
                return u, i
            return i
        return False

    def index_of_normalized_pair(self, c, d=None):
        r"""
        Return the index of the class `(c, d)` in the fixed list of
        representatives of `\mathbb(P)^1(R/N)`.

        INPUT:

        - ``c`` -- integral element of the corresponding number field, or a
          normalized :class:`MSymbol`

        - ``d`` -- (optional) when present, it must be an integral element of
          the number field such that `(c, d)` defines a normalized M-symbol of
          level `N`

        OUTPUT: ``i`` -- the index of `(c, d)` in the list

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 31)
            sage: N = k.ideal(5, a + 3)
            sage: P = P1NFList(N)
            sage: P.index_of_normalized_pair(1, 0)
            3
            sage: j = randint(0,len(P)-1)
            sage: P.index_of_normalized_pair(P[j])==j
            True
        """
        if d is None:
            try:
                c = MSymbol(self.__N, c)  # check that c is an MSymbol
            except ValueError:  # catch special case of wrong level
                raise ValueError("The MSymbol is of a different level")
            t, i = search(self.__list, c)
        else:
            t, i = search(self.__list, MSymbol(self.__N, c, d))
        if t:
            return i
        return False

    def lift_to_sl2_Ok(self, i):
        r"""
        Lift the `i`-th element of this :class:`P1NFList` to an element of `SL(2, R)`,
        where `R` is the ring of integers of the corresponding number field.

        INPUT:

        - ``i`` -- integer (index of the element to lift)

        OUTPUT:

        If the `i`-th element is `(c : d)`, the function returns a list of
        integral elements `[a, b, c', d']` that defines a `2\times 2` matrix with
        determinant 1 and such that `c=c'` (mod `N`) and `d=d'` (mod `N`).

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^2 + 23)
            sage: N = k.ideal(3)
            sage: P = P1NFList(N)
            sage: len(P)
            16
            sage: P[5]
            M-symbol (1/2*a + 1/2: -a) of level Fractional ideal (3)
            sage: P.lift_to_sl2_Ok(5)
            [-a, 2*a - 2, 1/2*a + 1/2, -a]

        ::

            sage: Ok = k.ring_of_integers()
            sage: L = [Matrix(Ok, 2, P.lift_to_sl2_Ok(i)) for i in range(len(P))]
            sage: all(det(L[i]) == 1 for i in range(len(L)))
            True
        """
        return self[i].lift_to_sl2_Ok()

    def apply_S(self, i):
        r"""
        Applies the matrix `S` = [0, -1, 1, 0] to the `i`-th M-Symbol of the list.

        INPUT:

        - ``i`` -- integer

        OUTPUT:

        integer -- the index of the M-Symbol obtained by the right action of
        the matrix `S` = [0, -1, 1, 0] on the `i`-th M-Symbol.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(5, a + 1)
            sage: P = P1NFList(N)
            sage: j = P.apply_S(P.index_of_normalized_pair(1, 0))
            sage: P[j]
            M-symbol (0: 1) of level Fractional ideal (5, a + 1)

        We test that S has order 2:

        ::

            sage: j = randint(0,len(P)-1)
            sage: P.apply_S(P.apply_S(j))==j
            True
        """
        c, d = self.__list[i].tuple()
        t, j = search(self.__list, self.normalize(d, -c))
        return j

    def apply_TS(self, i):
        r"""
        Applies the matrix `TS` = [1, -1, 0, 1] to the `i`-th M-Symbol of the list.

        INPUT:

        - ``i`` -- integer

        OUTPUT:

        integer -- the index of the M-Symbol obtained by the right action of
        the matrix `TS` = [1, -1, 0, 1] on the `i`-th M-Symbol.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(5, a + 1)
            sage: P = P1NFList(N)
            sage: P.apply_TS(3)
            2

        We test that `TS` has order 3:

        ::

            sage: j = randint(0,len(P)-1)
            sage: P.apply_TS(P.apply_TS(P.apply_TS(j)))==j
            True
        """
        c, d = self.__list[i].tuple()
        t, j = search(self.__list, self.normalize(c + d, -c))
        return j

    def apply_T_alpha(self, i, alpha=1):
        r"""
        Applies the matrix `T_{alpha}` = [1, `alpha`, 0, 1] to the `i`-th
        M-Symbol of the list.

        INPUT:

        - ``i`` -- integer

        - ``alpha`` -- (default: 1) element of the corresponding ring of integers

        OUTPUT:

        integer -- the index of the M-Symbol obtained by the right action of
        the matrix `T_{alpha}` = [1, `alpha`, 0, 1] on the `i`-th M-Symbol.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(5, a + 1)
            sage: P = P1NFList(N)
            sage: P.apply_T_alpha(4, a^ 2 - 2)
            3

        We test that `T_a*T_b = T_{(a+b)}`:

        ::

            sage: P.apply_T_alpha(3, a^2 - 2)==P.apply_T_alpha(P.apply_T_alpha(3,a^2),-2)
            True
        """
        c, d = self.__list[i].tuple()
        t, j = search(self.__list, self.normalize(c, alpha*c + d))
        return j

    def apply_J_epsilon(self, i, e1, e2=1):
        r"""
        Apply the matrix `J_{\epsilon}` = [e1, 0, 0, e2] to the `i`-th
        M-Symbol of the list.

        e1, e2 are units of the underlying number field.

        INPUT:

        - ``i`` -- integer

        - ``e1`` -- unit

        - ``e2`` -- unit (default: 1)

        OUTPUT:

        integer -- the index of the M-Symbol obtained by the right action of
        the matrix `J_{\epsilon}` = [e1, 0, 0, e2] on the `i`-th M-Symbol.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: k.<a> = NumberField(x^3 + 11)
            sage: N = k.ideal(5, a + 1)
            sage: P = P1NFList(N)
            sage: u = k.unit_group().gens_values(); u
            [-1, 2*a^2 + 4*a - 1]
            sage: P.apply_J_epsilon(4, -1)
            2
            sage: P.apply_J_epsilon(4, u[0], u[1])
            1

        ::

            sage: k.<a> = NumberField(x^4 + 13*x - 7)
            sage: N = k.ideal(a + 1)
            sage: P = P1NFList(N)
            sage: u = k.unit_group().gens_values(); u
            [-1, -a^3 - a^2 - a - 12, -a^3 - 3*a^2 + 1]
            sage: P.apply_J_epsilon(3, u[2]^2)==P.apply_J_epsilon(P.apply_J_epsilon(3, u[2]),u[2])
            True
        """
        c, d = self.__list[i].tuple()
        t, j = search(self.__list, self.normalize(c*e1, d*e2))
        return j


# *************************************************************************
#  Global functions:
#    - p1NFList --compute list of M-symbols
#    - lift_to_sl2_Ok
#    - make_coprime -- need it for ``lift_to_sl2_Ok``
#    - psi -- useful to check cardinality of the M-symbols list
# *************************************************************************

def p1NFlist(N):
    r"""
    Return a list of the normalized elements of `\mathbb{P}^1(R/N)`, where
    `N` is an integral ideal.

    INPUT:

    - ``N`` -- integral ideal (the level or modulus)

    EXAMPLES::

        sage: x = polygen(QQ, 'x')
        sage: k.<a> = NumberField(x^2 + 23)
        sage: N = k.ideal(3)
        sage: from sage.modular.modsym.p1list_nf import p1NFlist, psi
        sage: len(p1NFlist(N))==psi(N)
        True
    """
    k = N.number_field()

    L = [MSymbol(N, k(0), k(1), check=False)]
    # N.residues() = iterator through the residues mod N
    L = L + [MSymbol(N, k(1), r, check=False) for r in N.residues()]

    from sage.arith.misc import divisors
    for D in divisors(N):
        if not D.is_trivial() and D != N:
            # we find Dp ideal coprime to N, in inverse class to D
            if D.is_principal():
                Dp = k.ideal(1)
                c = D.gens_reduced()[0]
            else:
                it = k.primes_of_degree_one_iter()
                Dp = next(it)
                while not Dp.is_coprime(N) or not (Dp*D).is_principal():
                    Dp = next(it)
                c = (D*Dp).gens_reduced()[0]
            # now we find all the (c,d)'s which have associated divisor D
            I = D + N/D
            for d in (N/D).residues():
                if I.is_coprime(d):
                    M = D.prime_to_idealM_part(N/D)
                    u = (Dp*M).element_1_mod(N/D)
                    d1 = u*d + (1-u)
                    L.append(MSymbol(N, c, d1, check=False).normalize())
    return L


def lift_to_sl2_Ok(N, c, d):
    r"""
    Lift a pair (c, d) to an element of `SL(2, O_k)`, where `O_k` is the ring
    of integers of the corresponding number field.

    INPUT:

    - ``N`` -- number field ideal

    - ``c`` -- integral element of the number field

    - ``d`` -- integral element of the number field

    OUTPUT:

    A list `[a, b, c', d']` of integral elements that are the entries of
    a `2\times 2` matrix with determinant 1. The lower two entries are congruent to
    `c`, `d` modulo the ideal `N`.

    EXAMPLES::

        sage: from sage.modular.modsym.p1list_nf import lift_to_sl2_Ok
        sage: x = polygen(QQ, 'x')
        sage: k.<a> = NumberField(x^2 + 23)
        sage: Ok = k.ring_of_integers()
        sage: N = k.ideal(3)
        sage: M = Matrix(Ok, 2, lift_to_sl2_Ok(N, 1, a))
        sage: det(M)
        1
        sage: M = Matrix(Ok, 2, lift_to_sl2_Ok(N, 0, a))
        sage: det(M)
        1
        sage: (M[1][0] in N) and (M[1][1] - a in N)
        True
        sage: M = Matrix(Ok, 2, lift_to_sl2_Ok(N, 0, 0))
        Traceback (most recent call last):
        ...
        ValueError: Cannot lift (0, 0) to an element of Sl2(Ok).

    ::

        sage: k.<a> = NumberField(x^3 + 11)
        sage: Ok = k.ring_of_integers()
        sage: N = k.ideal(3, a - 1)
        sage: M = Matrix(Ok, 2, lift_to_sl2_Ok(N, 2*a, 0))
        sage: det(M)
        1
        sage: (M[1][0] - 2*a in N) and (M[1][1] in N)
        True
        sage: M = Matrix(Ok, 2, lift_to_sl2_Ok(N, 4*a^2, a + 1))
        sage: det(M)
        1
        sage: (M[1][0] - 4*a^2 in N) and (M[1][1] - (a+1) in N)
        True

    ::

        sage: k.<a> = NumberField(x^4 - x^3 -21*x^2 + 17*x + 133)
        sage: Ok = k.ring_of_integers()
        sage: N = k.ideal(7, a)
        sage: M = Matrix(Ok, 2, lift_to_sl2_Ok(N, 0, a^2 - 1))
        sage: det(M)
        1
        sage: (M[1][0] in N) and (M[1][1] - (a^2-1) in N)
        True
        sage: M = Matrix(Ok, 2, lift_to_sl2_Ok(N, 0, 7))
        Traceback (most recent call last):
        ...
        ValueError: <0> + <7> and the Fractional ideal (7, -4/7*a^3 + 13/7*a^2 + 39/7*a - 19) are not coprime.
    """
    k = N.number_field()
    # check the input
    if c.is_zero() and d.is_zero():
        raise ValueError("Cannot lift (%s, %s) to an element of Sl2(Ok)." % (c, d))
    if not N.is_coprime(k.ideal(c, d)):
        raise ValueError("<%s> + <%s> and the %s are not coprime." % (c, d, N))
    # a few special cases
    if c - 1 in N:
        return [k(0), k(-1), 1, d]
    if d - 1 in N:
        return [k(1), k(0), c, 1]
    if c.is_zero():  # and d!=1, so won't happen for normalized M-symbols (c: d)
        it = k.primes_of_degree_one_iter()
        q = k.ideal(1)
        while not (q.is_coprime(d) and (q*N).is_principal()):
            q = next(it)
        m = (q*N).gens_reduced()[0]
        B = k.ideal(m).element_1_mod(k.ideal(d))
        return [(1-B)/d, -B/m, m, d]
    if d.is_zero():  # and c!=1, so won't happen for normalized M-symbols (c: d)
        it = k.primes_of_degree_one_iter()
        q = k.ideal(1)
        while not (q.is_coprime(c) and (q*N).is_principal()):
            q = next(it)
        m = (q*N).gens_reduced()[0]
        B = k.ideal(c).element_1_mod(k.ideal(m))
        return [(1-B)/m, -B/c, c, m]

    c, d = make_coprime(N, c, d)

    B = k.ideal(c).element_1_mod(k.ideal(d))
    b = -B/c
    a = (1-B)/d
    return [a, b, c, d]


def make_coprime(N, c, d):
    r"""
    Return (c, d') so d' is congruent to d modulo N, and such that c and d' are
    coprime (`\langle c\rangle + \langle d'\rangle = R`).

    INPUT:

    - ``N`` -- number field ideal

    - ``c`` -- integral element of the number field

    - ``d`` -- integral element of the number field

    OUTPUT:

    A pair `(c, d')` where `c`, `d'` are integral elements of the corresponding
    number field, with `d'` congruent to `d` mod `N`, and such that `\langle c\rangle + \langle d'\rangle = R`
    (`R` being the corresponding ring of integers).

    EXAMPLES::

        sage: from sage.modular.modsym.p1list_nf import make_coprime
        sage: x = polygen(QQ, 'x')
        sage: k.<a> = NumberField(x^2 + 23)
        sage: N = k.ideal(3, a - 1)
        sage: c = 2*a; d = a + 1
        sage: N.is_coprime(k.ideal(c, d))
        True
        sage: k.ideal(c).is_coprime(d)
        False
        sage: c, dp = make_coprime(N, c, d)
        sage: k.ideal(c).is_coprime(dp)
        True
    """
    k = N.number_field()
    if k.ideal(c).is_coprime(d):
        return c, d
    q = k.ideal(c).prime_to_idealM_part(d)
    it = k.primes_of_degree_one_iter()
    r = k.ideal(1)
    qN = q*N
    while not (r.is_coprime(c) and (r*qN).is_principal()):
        r = next(it)
    m = (r*qN).gens_reduced()[0]
    d1 = d + m
    return c, d1


def psi(N):
    r"""
    The index `[\Gamma : \Gamma_0(N)]`, where `\Gamma = GL(2, R)` for `R` the
    corresponding ring of integers, and `\Gamma_0(N)` standard congruence
    subgroup.

    EXAMPLES::

        sage: from sage.modular.modsym.p1list_nf import psi
        sage: x = polygen(QQ, 'x')
        sage: k.<a> = NumberField(x^2 + 23)
        sage: N = k.ideal(3, a - 1)
        sage: psi(N)
        4

    ::

        sage: k.<a> = NumberField(x^2 + 23)
        sage: N = k.ideal(5)
        sage: psi(N)
        26
    """
    if not N.is_integral():
        raise ValueError("psi only defined for integral ideals")

    from sage.misc.misc_c import prod
    return prod([(np + 1) * np**(e - 1)
                 for np, e in [(p.absolute_norm(), e)
                               for p, e in N.factor()]])



#########################################################################
#       Copyright (C) 2004--2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
#########################################################################

from sage.modular.modform.constructor import ModularForms, CuspForms, EisensteinForms, Newforms, Newform

from sage.modular.modform.eis_series import eisenstein_series_qexp, eisenstein_series_lseries

from sage.modular.modform.half_integral import half_integral_weight_modform_basis

from sage.modular.modform.theta import theta_qexp, theta2_qexp

from sage.misc.lazy_import import lazy_import

lazy_import('sage.modular.modform.j_invariant', 'j_invariant_qexp')
lazy_import('sage.modular.modform.vm_basis', ['victor_miller_basis', 'delta_qexp'])

from sage.modular.modform.hecke_operator_on_qexp import hecke_operator_on_qexp, hecke_operator_on_basis

from sage.modular.modform.numerical import NumericalEigenforms as numerical_eigenforms

from sage.modular.modform.element import delta_lseries

from sage.modular.modform.ring import ModularFormsRing
del lazy_import



r"""
Mixed Differential Forms

Let `M` and `N` be differentiable manifolds and `\varphi : M \longrightarrow N`
a differentiable map. A *mixed differential form along* `\varphi` is an element
of the graded algebra represented by
:class:`~sage.manifolds.differentiable.mixed_form_algebra.MixedFormAlgebra`.
Its homogeneous components consist of differential forms along `\varphi`. Mixed
forms are useful to represent characteristic classes and perform computations
of such.

AUTHORS:

- Michael Jung (2019) : initial version
"""
# *****************************************************************************
#       Copyright (C) 2019 Michael Jung <micjung@uni-potsdam.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.rings.integer import Integer
from sage.structure.element import AlgebraElement, ModuleElementWithMutability


class MixedForm(AlgebraElement, ModuleElementWithMutability):
    r"""
    An instance of this class is a mixed form along some differentiable map
    `\varphi: M \to N` between two differentiable manifolds `M` and `N`. More
    precisely, a mixed form `a` along `\varphi: M \to N` can be considered as a
    differentiable map

    .. MATH::

        a: M \longrightarrow \bigoplus^n_{k=0} T^{(0,k)}N,

    where `T^{(0,k)}` denotes the tensor bundle of type `(0,k)`, `\bigoplus`
    the Whitney sum and `n` the dimension of `N`, such that

    .. MATH::

        \forall x\in M, \quad a(x) \in \bigoplus^n_{k=0} \Lambda^k\left( T_{\varphi(x)}^* N \right),

    where `\Lambda^k(T^*_{\varphi(x)} N)` is the `k`-th exterior power of the
    dual of the tangent space `T_{\varphi(x)} N`. Thus, a mixed differential
    form `a` consists of homogeneous components `a_i`, `i=0,1, \dots, n`, where
    the `i`-th homogeneous component represents a differential form of
    degree `i`.

    The standard case of a mixed form *on* `M` corresponds to `M=N` with
    `\varphi = \mathrm{Id}_M`.

    INPUT:

    - ``parent`` -- graded algebra of mixed forms represented by
      :class:`~sage.manifolds.differentiable.mixed_form_algebra.MixedFormAlgebra`
      where the mixed form ``self`` shall belong to
    - ``comp`` -- (default: ``None``) homogeneous components of the mixed form
      as a list; if none is provided, the components are set to innocent unnamed
      differential forms
    - ``name`` -- (default: ``None``) name given to the mixed form
    - ``latex_name`` -- (default: ``None``) LaTeX symbol to denote the
      mixed form; if none is provided, the LaTeX symbol is set to ``name``

    EXAMPLES:

    Initialize a mixed form on a 2-dimensional parallelizable differentiable
    manifold::

        sage: M = Manifold(2, 'M')
        sage: c_xy.<x,y> = M.chart()
        sage: e_xy = c_xy.frame()
        sage: A = M.mixed_form(name='A'); A
        Mixed differential form A on the 2-dimensional differentiable manifold M
        sage: A.parent()
        Graded algebra Omega^*(M) of mixed differential forms on the
         2-dimensional differentiable manifold M

    The default way to specify the `i`-th homogeneous component
    of a mixed form is by accessing it via ``A[i]`` or using :meth:`set_comp`::

        sage: A = M.mixed_form(name='A')
        sage: A[0].set_expr(x) # scalar field
        sage: A.set_comp(1)[0] = y*x
        sage: A.set_comp(2)[0,1] = y^2*x
        sage: A.display() # display names
        A = A_0 + A_1 + A_2
        sage: A.display_expansion() # display expansion in basis
        A = x + x*y dx + x*y^2 dx∧dy

    Another way to define the homogeneous components is using predefined
    differential forms::

        sage: f = M.scalar_field(x, name='f'); f
        Scalar field f on the 2-dimensional differentiable manifold M
        sage: omega = M.diff_form(1, name='omega'); omega
        1-form omega on the 2-dimensional differentiable manifold M
        sage: omega[e_xy,0] = y*x; omega.display()
        omega = x*y dx
        sage: eta = M.diff_form(2, name='eta'); eta
        2-form eta on the 2-dimensional differentiable manifold M
        sage: eta[e_xy,0,1] = y^2*x; eta.display()
        eta = x*y^2 dx∧dy

    The components of a mixed form ``B`` can then be set as follows::

        sage: B = M.mixed_form(name='B')
        sage: B[:] = [f, omega, eta]; B.display() # display names
        B = f + omega + eta
        sage: B.display_expansion() # display in coordinates
        B = x + x*y dx + x*y^2 dx∧dy
        sage: B[0]
        Scalar field f on the 2-dimensional differentiable manifold M
        sage: B[1]
        1-form omega on the 2-dimensional differentiable manifold M
        sage: B[2]
        2-form eta on the 2-dimensional differentiable manifold M

    As we can see, the names are applied. However note that the differential
    forms are different instances::

        sage: f is B[0]
        False

    Alternatively, the components can be determined from scratch::

        sage: B = M.mixed_form([f, omega, eta], name='B')
        sage: B.display()
        B = f + omega + eta

    Mixed forms are elements of an algebra so they can be added, and multiplied
    via the wedge product::

        sage: C = x*A; C
        Mixed differential form x∧A on the 2-dimensional differentiable
         manifold M
        sage: C.display_expansion()
        x∧A = x^2 + x^2*y dx + x^2*y^2 dx∧dy
        sage: D = A+C; D
        Mixed differential form A+x∧A on the 2-dimensional differentiable
         manifold M
        sage: D.display_expansion()
        A+x∧A = x^2 + x + (x^2 + x)*y dx + (x^2 + x)*y^2 dx∧dy
        sage: E = A*C; E
        Mixed differential form A∧(x∧A) on the 2-dimensional differentiable
         manifold M
        sage: E.display_expansion()
        A∧(x∧A) = x^3 + 2*x^3*y dx + 2*x^3*y^2 dx∧dy

    Coercions are fully implemented::

        sage: F = omega*A
        sage: F.display_expansion()
        omega∧A = x^2*y dx
        sage: G = omega+A
        sage: G.display_expansion()
        omega+A = x + 2*x*y dx + x*y^2 dx∧dy

    Moreover, it is possible to compute the exterior derivative of a
    mixed form::

        sage: dA = A.exterior_derivative(); dA.display()
        dA = zero + dA_0 + dA_1
        sage: dA.display_expansion()
        dA = dx - x dx∧dy

    Initialize a mixed form on a 2-dimensional non-parallelizable differentiable
    manifold::

        sage: M = Manifold(2, 'M')
        sage: U = M.open_subset('U') ; V = M.open_subset('V')
        sage: M.declare_union(U,V)   # M is the union of U and V
        sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
        sage: transf = c_xy.transition_map(c_uv, (x+y, x-y),
        ....:                   intersection_name='W', restrictions1= x>0,
        ....:                   restrictions2= u+v>0)
        sage: inv = transf.inverse()
        sage: W = U.intersection(V)
        sage: e_xy = c_xy.frame(); e_uv = c_uv.frame() # define frames
        sage: A = M.mixed_form(name='A')
        sage: A[0].set_expr(x, c_xy)
        sage: A[0].display()
        A_0: M → ℝ
        on U: (x, y) ↦ x
        on W: (u, v) ↦ 1/2*u + 1/2*v
        sage: A[1][0] = y*x; A[1].display(e_xy)
        A_1 = x*y dx
        sage: A[2][e_uv,0,1] = u*v^2; A[2].display(e_uv)
        A_2 = u*v^2 du∧dv
        sage: A.add_comp_by_continuation(e_uv, W, c_uv)
        sage: A.display_expansion(e_uv)
        A = 1/2*u + 1/2*v + (1/8*u^2 - 1/8*v^2) du + (1/8*u^2 - 1/8*v^2) dv + u*v^2 du∧dv
        sage: A.add_comp_by_continuation(e_xy, W, c_xy)
        sage: A.display_expansion(e_xy)
        A = x + x*y dx + (-2*x^3 + 2*x^2*y + 2*x*y^2 - 2*y^3) dx∧dy

    Since zero and one are special elements, their components cannot be
    changed::

        sage: z = M.mixed_form_algebra().zero()
        sage: z[0] = 1
        Traceback (most recent call last):
        ...
        ValueError: the components of an immutable element cannot be changed
        sage: one = M.mixed_form_algebra().one()
        sage: one[0] = 0
        Traceback (most recent call last):
        ...
        ValueError: the components of an immutable element cannot be changed
    """
    def __init__(self, parent, name=None, latex_name=None):
        r"""
        Construct a mixed form.

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U') ; V = M.open_subset('V')
            sage: M.declare_union(U,V)
            sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
            sage: transf = c_xy.transition_map(c_uv, (x+y, x-y),
            ....:                   intersection_name='W', restrictions1= x>0,
            ....:                   restrictions2= u+v>0)
            sage: inv = transf.inverse()
            sage: W = U.intersection(V)
            sage: e_xy = c_xy.frame(); e_uv = c_uv.frame()
            sage: omega = M.diff_form(1, name='omega')
            sage: omega[e_xy,0] = y*x
            sage: omega.add_comp_by_continuation(e_uv, W, c_uv)
            sage: eta = M.diff_form(2, name='eta')
            sage: eta[e_uv,0,1] = u*v^2
            sage: eta.add_comp_by_continuation(e_xy, W, c_xy)
            sage: A = M.mixed_form_algebra()
            sage: F = A([x, omega, eta], name='F')
            sage: TestSuite(F).run(skip='_test_pickling')
        """
        if parent is None:
            raise ValueError("a parent must be provided")
        # Add this element (instance) to the set of parent algebra:
        AlgebraElement.__init__(self, parent)
        # Get the underlying vector field module:
        vmodule = parent._vmodule
        # Define attributes:
        self._vmodule = vmodule
        self._dest_map = vmodule._dest_map
        self._domain = vmodule._domain
        self._ambient_domain = vmodule._ambient_domain
        self._max_deg = vmodule._ambient_domain.dim()
        self._is_zero = False  # a priori, may be changed below or via
        # method __bool__()
        self._comp = None  # initialized on demand; see _init_comp
        # Set names:
        self._name = name
        if latex_name is None:
            self._latex_name = self._name
        else:
            self._latex_name = latex_name

    def _init_comp(self):
        r"""
        Initialize homogeneous components of ``self``.

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: A = M.mixed_form(name='A')
            sage: A._comp is None
            True
            sage: A._init_comp()
            sage: A._comp
            [Scalar field A_0 on the 2-dimensional differentiable manifold M,
             1-form A_1 on the 2-dimensional differentiable manifold M,
             2-form A_2 on the 2-dimensional differentiable manifold M]
        """
        self._comp = []
        for i in self.irange():
            comp_name, comp_latex_name = None, None
            if self._name is not None:
                comp_name = f"{self._name}_{i}"
            if self._latex_name is not None:
                comp_latex_name = '{' + self._latex_name + '}_{' + str(i) + '}'
            diff_form = self._domain.diff_form
            self._comp.append(diff_form(i, name=comp_name,
                                        latex_name=comp_latex_name))

    def _repr_(self):
        r"""
        String representation of ``self``.

        TESTS::

            sage: M = Manifold(3, 'M')
            sage: F = M.mixed_form(name='F')
            sage: F._repr_()
            'Mixed differential form F on the 3-dimensional differentiable
             manifold M'
            sage: repr(F)  # indirect doctest
            'Mixed differential form F on the 3-dimensional differentiable
             manifold M'

        Check whether :issue:`31784` is fixed::

            sage: E3 = EuclideanSpace(3)
            sage: S2 = E3.sphere()
            sage: iota = S2.embedding()
            sage: Omega = S2.mixed_form_algebra(dest_map=iota)
            sage: Omega(1)
            Mixed differential form one along the 2-sphere S^2 of radius 1
             smoothly embedded in the Euclidean space E^3 with values on the
             Euclidean space E^3 via the map iota
        """
        desc = "Mixed differential form "
        if self._name is not None:
            desc += self._name + " "
        if self._dest_map is self._domain.identity_map():
            desc += f"on the {self._domain}"
        else:
            desc += f"along the {self._domain} with values on the {self._ambient_domain} "
            if self._dest_map._name is None:
                dm_name = "unnamed map"
            else:
                dm_name = self._dest_map._name
            desc += "via the map " + dm_name
        return desc

    def _latex_(self):
        r"""
        Return a LaTeX representation of the object.

        TESTS::

            sage: M = Manifold(3, 'M', latex_name=r'\mathcal{M}')
            sage: omega = M.mixed_form(name='omega', latex_name=r'\omega')
            sage: omega._latex_()
            '\\omega'
            sage: latex(omega)  # indirect doctest
            \omega
        """
        if self._name is None:
            return r'\text{' + repr(self) + r'}'
        return self._latex_name

    def _new_instance(self, name=None, latex_name=None):
        r"""
        Return a new instance of ``self``.

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: F = M.mixed_form(name='F')
            sage: F1 = F._new_instance(); F1
            Mixed differential form on the 2-dimensional differentiable
             manifold M
            sage: type(F1) == type(F)
            True
            sage: F1.parent() is F.parent()
            True
        """
        return type(self)(self.parent(), name=name, latex_name=latex_name)

    def display_expansion(self, frame=None, chart=None, from_chart=None):
        r"""
        Display the expansion in a particular basis and chart of mixed forms.

        The output is either text-formatted (console mode) or LaTeX-formatted
        (notebook mode).

        INPUT:

        - ``frame`` -- (default: ``None``) vector frame with respect to
          which the mixed form is expanded; if ``None``, only the names
          of the components are displayed
        - ``chart`` -- (default: ``None``) chart with respect to which the
          components of the mixed form in the selected frame are expressed;
          if ``None``, the default chart of the vector frame domain is assumed

        EXAMPLES:

        Display the expansion of a mixed form on a 2-dimensional
        non-parallelizable differentiable manifold::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U') ; V = M.open_subset('V')
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
            sage: transf = c_xy.transition_map(c_uv, (x-y, x+y),
            ....:                   intersection_name='W', restrictions1= x>0,
            ....:                   restrictions2= u+v>0)
            sage: inv = transf.inverse()
            sage: W = U.intersection(V)
            sage: e_xy = c_xy.frame(); e_uv = c_uv.frame() # define frames
            sage: omega = M.diff_form(1, name='omega')
            sage: omega[e_xy,0] = x; omega.display(e_xy)
            omega = x dx
            sage: omega.add_comp_by_continuation(e_uv, W, c_uv) # continuation onto M
            sage: eta = M.diff_form(2, name='eta')
            sage: eta[e_uv,0,1] = u*v; eta.display(e_uv)
            eta = u*v du∧dv
            sage: eta.add_comp_by_continuation(e_xy, W, c_xy) # continuation onto M
            sage: F = M.mixed_form([0, omega, eta], name='F'); F
            Mixed differential form F on the 2-dimensional differentiable
             manifold M
            sage: F.display() # display names of homogeneous components
            F = zero + omega + eta
            sage: F.display_expansion(e_uv)
            F = (1/4*u + 1/4*v) du + (1/4*u + 1/4*v) dv + u*v du∧dv
            sage: F.display_expansion(e_xy)
            F = x dx + (2*x^2 - 2*y^2) dx∧dy
        """
        from sage.misc.latex import latex
        from sage.tensor.modules.format_utilities import FormattedExpansion, is_atomic
        from sage.typeset.unicode_characters import unicode_wedge
        # In case, no frame is given:
        if frame is None:
            frame = self._domain._def_frame
        # In case, no chart is given:
        if chart is None:
            chart = frame._chart
        # Check names:
        if self._name is not None:
            resu_txt = self._name + " = "
        else:
            resu_txt = ""
        if self._latex_name is not None:
            resu_latex = self._latex_name + r" = "
        else:
            resu_latex = ""
        # Get terms
        terms_txt = []
        terms_latex = []
        # Scalar field term:
        if not self[0].is_trivial_zero():
            terms_txt.append(repr(self[0].expr(chart, from_chart)))
            terms_latex.append(latex(self[0].expr(chart, from_chart)))
        # Differential form terms:
        for j in self.irange(1):
            rst = self[j].restrict(frame._domain, dest_map=frame._dest_map)
            basis, format_spec = rst._preparse_display(basis=frame,
                                                       format_spec=chart)
            cobasis = basis.dual_basis()
            comp = rst.comp(basis)
            for ind in comp.non_redundant_index_generator():
                ind_arg = ind + (format_spec,)
                coef = comp[ind_arg]
                # Check whether the coefficient is zero, preferably via
                # the fast method is_trivial_zero():
                if hasattr(coef, 'is_trivial_zero'):
                    zero_coef = coef.is_trivial_zero()
                else:
                    zero_coef = coef == 0
                if not zero_coef:
                    bases_txt = []
                    bases_latex = []
                    for k in range(rst._tensor_rank):
                        bases_txt.append(cobasis[ind[k]]._name)
                        bases_latex.append(latex(cobasis[ind[k]]))
                    basis_term_txt = unicode_wedge.join(bases_txt)
                    basis_term_latex = r"\wedge ".join(bases_latex)
                    coef_txt = repr(coef)
                    if coef_txt == "1":
                        terms_txt.append(basis_term_txt)
                        terms_latex.append(basis_term_latex)
                    elif coef_txt == "-1":
                        terms_txt.append("-" + basis_term_txt)
                        terms_latex.append("-" + basis_term_latex)
                    else:
                        coef_latex = latex(coef)
                        if is_atomic(coef_txt):
                            terms_txt.append(coef_txt + " " + basis_term_txt)
                        else:
                            terms_txt.append("(" + coef_txt + ") " +
                                             basis_term_txt)
                        if is_atomic(coef_latex):
                            terms_latex.append(coef_latex + basis_term_latex)
                        else:
                            terms_latex.append(r"\left(" + coef_latex +
                                               r"\right)" + basis_term_latex)
        if not terms_txt:
            resu_txt += "0"
        else:
            resu_txt += terms_txt[0]
            for term in terms_txt[1:]:
                if term[0] == "-":
                    resu_txt += " - " + term[1:]
                else:
                    resu_txt += " + " + term
        if not terms_latex:
            resu_latex += "0"
        else:
            resu_latex += terms_latex[0]
            for term in terms_latex[1:]:
                if term[0] == "-":
                    resu_latex += term
                else:
                    resu_latex += "+" + term
        return FormattedExpansion(resu_txt, resu_latex)

    disp_exp = display_expansion
    display_exp = display_expansion

    def display(self):
        r"""
        Display the homogeneous components of the mixed form.

        The output is either text-formatted (console mode) or LaTeX-formatted
        (notebook mode).

        EXAMPLES::

            sage: M = Manifold(2, 'M')
            sage: f = M.scalar_field(name='f')
            sage: omega = M.diff_form(1, name='omega')
            sage: eta = M.diff_form(2, name='eta')
            sage: F = M.mixed_form([f, omega, eta], name='F'); F
            Mixed differential form F on the 2-dimensional differentiable
             manifold M
            sage: F.display() # display names of homogeneous components
            F = f + omega + eta
        """
        from sage.misc.latex import latex
        from sage.tensor.modules.format_utilities import FormattedExpansion
        # Mixed form name:
        if self._name is not None:
            resu_txt = self._name + " = "
        else:
            resu_txt = ""
        if self._latex_name is not None:
            resu_latex = self._latex_name + r" = "
        else:
            resu_latex = ""
        # Scalar field:
        if self[0]._name is None:
            resu_txt += "(unnamed scalar field) "
        else:
            resu_txt += self[0]._name
        if self[0]._latex_name is None:
            resu_latex += r"\text{(unnamed scalar field)}"
        else:
            resu_latex += latex(self[0])
        # Differential forms:
        for j in self.irange(1):
            if self[j]._name is None:
                resu_txt += " + (unnamed " + str(j) + "-form)"
            else:
                resu_txt += " + " + self[j]._name
            if self[j]._latex_name is None:
                resu_latex += r"+\text{(unnamed " + str(j) + r"-form)}"
            else:
                resu_latex += r"+" + latex(self[j])
        return FormattedExpansion(resu_txt, resu_latex)

    disp = display

    def set_name(self, name=None, latex_name=None, apply_to_comp=True):
        r"""
        Redefine the string and LaTeX representation of the object.

        INPUT:

        - ``name`` -- (default: ``None``) name given to the mixed form
        - ``latex_name`` -- (default: ``None``) LaTeX symbol to denote the
          mixed form; if none is provided, the LaTeX symbol is set to ``name``
        - ``apply_to_comp`` -- boolean (default: ``True``); if ``True`` all homogeneous
          components will be renamed accordingly. If ``False`` only the mixed
          form will be renamed

        EXAMPLES:

        Rename a mixed form::

            sage: M = Manifold(4, 'M')
            sage: F = M.mixed_form(name='dummy', latex_name=r'\ugly'); F
            Mixed differential form dummy on the 4-dimensional differentiable
             manifold M
            sage: latex(F)
            \ugly
            sage: F.set_name(name='F', latex_name=r'\mathcal{F}'); F
            Mixed differential form F on the 4-dimensional differentiable
             manifold M
            sage: latex(F)
            \mathcal{F}

        If not stated otherwise, all homogeneous components are renamed
        accordingly::

            sage: F.display()
            F = F_0 + F_1 + F_2 + F_3 + F_4

        Setting the argument ``set_all`` to ``False`` prevents the renaming
        in the homogeneous components::

            sage: F.set_name(name='eta', latex_name=r'\eta', apply_to_comp=False)
            sage: F.display()
            eta = F_0 + F_1 + F_2 + F_3 + F_4

        To rename a homogeneous component individually, we simply access the
        homogeneous component and use its
        :meth:`~sage.manifolds.differentiable.tensorfield.set_name` method::

            sage: F[0].set_name(name='g'); F.display()
            eta = g + F_1 + F_2 + F_3 + F_4
        """
        if self.is_immutable():
            raise ValueError("the name of an immutable element "
                             "cannot be changed")
        if name is not None:
            self._name = name
            if latex_name is None:
                self._latex_name = self._name
        if latex_name is not None:
            self._latex_name = latex_name
        if apply_to_comp:
            for i in self.irange():
                comp_name, comp_latex_name = None, None
                if self._name is not None:
                    comp_name = f"{self._name}_{i}"
                if self._latex_name is not None:
                    comp_latex_name = '{' + self._latex_name + '}_{' + str(i) + '}'
                self[i].set_name(name=comp_name, latex_name=comp_latex_name)

    def __bool__(self):
        r"""
        Return ``True`` if ``self`` is nonzero and ``False`` otherwise.

        This method is called by :meth:`is_zero`.

        EXAMPLES:

        Mixed form defined by parts on a 2-dimensional manifold::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U')
            sage: c_xy.<x,y> = U.chart()
            sage: V = M.open_subset('V')
            sage: c_uv.<u,v> = V.chart()
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: F = M.mixed_form(name='F')
            sage: FU = U.mixed_form(name='F')
            sage: FV = V.mixed_form(name='F')
            sage: FU[:] = [0,0,0]
            sage: FV[:] = [0,0,0]
            sage: F.set_restriction(FU)
            sage: F.set_restriction(FV)
            sage: bool(F)
            False
            sage: F.is_zero()  # indirect doctest
            True
            sage: FU[0] = 1
            sage: F.set_restriction(FU)
            sage: bool(F)
            True
            sage: F.is_zero()  # indirect doctest
            False
        """
        if self._is_zero:
            return False
        if any(bool(form) for form in self):
            self._is_zero = False
            return True
        self._is_zero = True
        return False

    def _richcmp_(self, other, op):
        r"""
        Comparison method for sage objects.

        INPUT:

        - ``other`` -- a mixed form
        - ``op`` -- comparison operator for which ``self`` and ``other`` shall
          be compared with

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U') ; V = M.open_subset('V')
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
            sage: xy_to_uv = c_xy.transition_map(c_uv, (x+y, x-y),
            ....:                    intersection_name='W', restrictions1= x>0,
            ....:                    restrictions2= u+v>0)
            sage: uv_to_xy = xy_to_uv.inverse()
            sage: W = U.intersection(V)
            sage: e_xy = c_xy.frame(); e_uv = c_uv.frame()
            sage: f = M.scalar_field(x, name='f')
            sage: f.add_expr_by_continuation(c_uv, W)
            sage: eta = M.diff_form(1, name='eta')
            sage: eta[e_xy,0] = x+y
            sage: eta.add_comp_by_continuation(e_uv, W, c_uv)
            sage: F = M.mixed_form([f, eta, 0])
            sage: F == F
            True
            sage: F == F.copy()
            True
            sage: G = M.mixed_form()
            sage: G.set_restriction(F.restrict(U))
            sage: F == G  # False since G has not been defined on V
            False
            sage: G.set_restriction(F.restrict(V))
            sage: F == G  # ``True`` now
            True
            sage: H = M.mixed_form([f, 0, 0])
            sage: F != H  # this is fixed by issue #30108
            True
            sage: F.parent().zero() == 0
            True
        """
        from sage.structure.richcmp import op_EQ, op_NE
        if op == op_NE:
            return not self == other
        if op == op_EQ:
            # Compare all elements separately:
            return all(self[j] == other[j] for j in self.irange())
        # Fall back on default implementation:
        return super()._richcmp_(self, other, op)

    def _add_(self, other):
        r"""
        Addition of two mixed forms.

        INPUT:

        - ``other`` -- a mixed form, in the same algebra as ``self``

        OUTPUT:

        - the mixed form resulting from the addition of ``self``
          and ``other``

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U') ; V = M.open_subset('V')
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
            sage: xy_to_uv = c_xy.transition_map(c_uv, (x+y, x-y),
            ....:                    intersection_name='W', restrictions1= x>0,
            ....:                    restrictions2= u+v>0)
            sage: uv_to_xy = xy_to_uv.inverse()
            sage: W = U.intersection(V)
            sage: e_xy = c_xy.frame(); e_uv = c_uv.frame()
            sage: f = M.scalar_field(x, name='f')
            sage: f.add_expr_by_continuation(c_uv, W) # continuation onto M
            sage: a = M.diff_form(1, name='a')
            sage: a[e_xy,0] = x
            sage: a.add_comp_by_continuation(e_uv, W, c_uv) # continuation onto M
            sage: A = M.mixed_form([f, a, 0], name='A')
            sage: g = M.scalar_field(u, name='g', chart=c_uv)
            sage: g.add_expr_by_continuation(c_xy, W) # continuation onto M
            sage: b = M.diff_form(1, name='b')
            sage: b[e_uv,1] = v
            sage: b.add_comp_by_continuation(e_xy, W, c_xy) # continuation onto M
            sage: B = M.mixed_form([g, b, 0], name='B')
            sage: C = A._add_(B); C
            Mixed differential form A+B on the 2-dimensional differentiable
             manifold M
            sage: A.display_expansion(e_uv)
            A = 1/2*u + 1/2*v + (1/4*u + 1/4*v) du + (1/4*u + 1/4*v) dv
            sage: B.display_expansion(e_xy)
            B = x + y + (x - y) dx + (-x + y) dy
            sage: C.display_expansion(e_xy)
            A+B = 2*x + y + (2*x - y) dx + (-x + y) dy
            sage: C.display_expansion(e_uv)
            A+B = 3/2*u + 1/2*v + (1/4*u + 1/4*v) du + (1/4*u + 5/4*v) dv
            sage: C == A + B  # indirect doctest
            True
            sage: Z = A.parent().zero(); Z
            Mixed differential form zero on the 2-dimensional differentiable
             manifold M
            sage: A._add_(Z) == A
            True
            sage: Z._add_(A) == A
            True
        """
        # Case zero:
        if self._is_zero:
            return other
        if other._is_zero:
            return self
        # Generic case:
        resu = self._new_instance()
        resu._comp = [self[j] + other[j] for j in self.irange()]
        # Compose name:
        if self._name is not None and other._name is not None:
            resu._name = self._name + '+' + other._name
        if self._latex_name is not None and other._latex_name is not None:
            resu._latex_name = self._latex_name + '+' + other._latex_name
        return resu

    def _sub_(self, other):
        r"""
        Subtraction of two mixed forms.

        INPUT:

        - ``other`` -- a mixed form, in the same algebra as ``self``

        OUTPUT:

        - the mixed form resulting from the subtraction of ``self``
          and ``other``

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U') ; V = M.open_subset('V')
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
            sage: xy_to_uv = c_xy.transition_map(c_uv, (x+y, x-y),
            ....:                    intersection_name='W', restrictions1= x>0,
            ....:                    restrictions2= u+v>0)
            sage: uv_to_xy = xy_to_uv.inverse()
            sage: W = U.intersection(V)
            sage: e_xy = c_xy.frame(); e_uv = c_uv.frame()
            sage: f = M.scalar_field(x, name='f')
            sage: f.add_expr_by_continuation(c_uv, W) # continuation onto M
            sage: a = M.diff_form(1, name='a')
            sage: a[e_xy,0] = x
            sage: a.add_comp_by_continuation(e_uv, W, c_uv) # continuation onto M
            sage: A = M.mixed_form([f, a, 0], name='A')
            sage: g = M.scalar_field(u, name='g', chart=c_uv)
            sage: g.add_expr_by_continuation(c_xy, W) # continuation onto M
            sage: b = M.diff_form(1, name='b')
            sage: b[e_uv,1] = v
            sage: b.add_comp_by_continuation(e_xy, W, c_xy) # continuation onto M
            sage: B = M.mixed_form([g, b, 0], name='B')
            sage: C = A._sub_(B); C
            Mixed differential form A-B on the 2-dimensional differentiable
             manifold M
            sage: A.display_expansion(e_uv)
            A = 1/2*u + 1/2*v + (1/4*u + 1/4*v) du + (1/4*u + 1/4*v) dv
            sage: B.display_expansion(e_xy)
            B = x + y + (x - y) dx + (-x + y) dy
            sage: C.display_expansion(e_xy)
            A-B = -y + y dx + (x - y) dy
            sage: C.display_expansion(e_uv)
            A-B = -1/2*u + 1/2*v + (1/4*u + 1/4*v) du + (1/4*u - 3/4*v) dv
            sage: C == A - B  # indirect doctest
            True
            sage: Z = A.parent().zero(); Z
            Mixed differential form zero on the 2-dimensional differentiable
             manifold M
            sage: A._sub_(Z) == A
            True
            sage: Z._sub_(A) == -A
            True
        """
        # Case zero:
        if self._is_zero:
            return -other
        if other._is_zero:
            return self
        # Generic case:
        resu = self._new_instance()
        resu._comp = [self[j] - other[j] for j in self.irange()]
        # Compose name:
        from sage.tensor.modules.format_utilities import is_atomic

        if self._name is not None and other._name is not None:
            sname = self._name
            oname = other._name
            if not is_atomic(oname):
                oname = '(' + oname + ')'
            resu._name = sname + '-' + oname
        if self._latex_name is not None and other._latex_name is not None:
            slname = self._latex_name
            olname = other._latex_name
            if not is_atomic(olname):
                olname = '(' + olname + ')'
            resu._latex_name = slname + '-' + olname
        return resu

    def wedge(self, other):
        r"""
        Wedge product on the graded algebra of mixed forms.

        More precisely, the wedge product is a bilinear map

        .. MATH::

            \wedge: \Omega^k(M,\varphi) \times \Omega^l(M,\varphi) \to \Omega^{k+l}(M,\varphi),

        where `\Omega^k(M,\varphi)` denotes the space of differential forms of
        degree `k` along `\varphi`. By bilinear extension, this induces a map

        .. MATH::

            \wedge: \Omega^*(M,\varphi) \times \Omega^*(M,\varphi) \to \Omega^*(M,\varphi) ``

        and equips `\Omega^*(M,\varphi)` with a multiplication such that it
        becomes a graded algebra.

        INPUT:

        - ``other`` -- mixed form in the same algebra as ``self``

        OUTPUT:

        - the mixed form resulting from the wedge product of ``self``
          with ``other``

        EXAMPLES:

        Initialize a mixed form on a 3-dimensional manifold::

            sage: M = Manifold(3, 'M')
            sage: c_xyz.<x,y,z> = M.chart()
            sage: f = M.scalar_field(x, name='f')
            sage: f.display()
            f: M → ℝ
               (x, y, z) ↦ x
            sage: g = M.scalar_field(y, name='g')
            sage: g.display()
            g: M → ℝ
               (x, y, z) ↦ y
            sage: omega = M.diff_form(1, name='omega')
            sage: omega[0] = x
            sage: omega.display()
            omega = x dx
            sage: eta = M.diff_form(1, name='eta')
            sage: eta[1] = y
            sage: eta.display()
            eta = y dy
            sage: mu = M.diff_form(2, name='mu')
            sage: mu[0,2] = z
            sage: mu.display()
            mu = z dx∧dz
            sage: A = M.mixed_form([f, omega, mu, 0], name='A')
            sage: A.display_expansion()
            A = x + x dx + z dx∧dz
            sage: B = M.mixed_form([g, eta, mu, 0], name='B')
            sage: B.display_expansion()
            B = y + y dy + z dx∧dz

        The wedge product of ``A`` and ``B`` yields::

            sage: C = A.wedge(B); C
            Mixed differential form A∧B on the 3-dimensional differentiable
             manifold M
            sage: C.display_expansion()
            A∧B = x*y + x*y dx + x*y dy + x*y dx∧dy + (x + y)*z dx∧dz - y*z dx∧dy∧dz
            sage: D = B.wedge(A); D # Don't even try, it's not commutative!
            Mixed differential form B∧A on the 3-dimensional differentiable
             manifold M
            sage: D.display_expansion() # I told you so!
            B∧A = x*y + x*y dx + x*y dy - x*y dx∧dy + (x + y)*z dx∧dz - y*z dx∧dy∧dz

        Alternatively, the multiplication symbol can be used::

            sage: A*B
            Mixed differential form A∧B on the 3-dimensional differentiable
             manifold M
            sage: A*B == C
            True

        Yet, the multiplication includes coercions::

            sage: E = x*A; E.display_expansion()
            x∧A = x^2 + x^2 dx + x*z dx∧dz
            sage: F = A*eta; F.display_expansion()
            A∧eta = x*y dy + x*y dx∧dy - y*z dx∧dy∧dz
        """
        # Case zero:
        if self._is_zero or other._is_zero:
            return self.parent().zero()
        # Case one:
        if self is self.parent().one():
            return other
        if other is self.parent().one():
            return self
        # Generic case:
        resu = self._new_instance()
        resu._comp = [sum(self[k].wedge(other[j - k]) for k in range(j + 1))
                      for j in self.irange()]
        # Compose name:
        from sage.tensor.modules.format_utilities import (
            format_mul_latex,
            format_mul_txt,
        )
        from sage.typeset.unicode_characters import unicode_wedge
        resu._name = format_mul_txt(self._name, unicode_wedge, other._name)
        resu._latex_name = format_mul_latex(self._latex_name, r'\wedge ',
                                            other._latex_name)
        return resu

    _mul_ = wedge

    def _lmul_(self, other):
        r"""
        Scalar multiplication operator: return ``number * self`` or
        ``self * number``.

        INPUT:

        - ``other`` -- an element of the symbolic ring

        OUTPUT:

        - the mixed form resulting from the wedge product of ``self``
          with ``other``

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: c_xy.<x,y> = M.chart()
            sage: omega = M.diff_form(1, name='omega')
            sage: omega[0] = y*x; omega.display()
            omega = x*y dx
            sage: F = M.mixed_form([0, omega, 0], name='F')
            sage: A = x*F*y; A
            Mixed differential form y∧(x∧F) on the 2-dimensional
             differentiable manifold M
            sage: A.display_expansion()
            y∧(x∧F) = x^2*y^2 dx
        """
        try:
            if other.is_trivial_zero():
                return self.parent().zero()
            if (other - 1).is_trivial_zero():
                return self
        except AttributeError:
            # in case base ring is not SR:
            if other == 0:
                return self.parent().zero()
            if other == 1:
                return self
        resu = self._new_instance()
        resu._comp = [other * form for form in self]
        # Compose name:
        from sage.misc.latex import latex
        from sage.tensor.modules.format_utilities import (
            format_mul_latex,
            format_mul_txt,
        )
        from sage.typeset.unicode_characters import unicode_wedge
        resu._name = format_mul_txt(repr(other), unicode_wedge, self._name)
        resu._latex_name = format_mul_latex(latex(other), r'\wedge ',
                                            self._latex_name)
        return resu

    @cached_method
    def exterior_derivative(self):
        r"""
        Compute the exterior derivative of ``self``.

        More precisely, the *exterior derivative* on `\Omega^k(M,\varphi)` is a
        linear map

        .. MATH::

            \mathrm{d}_{k} : \Omega^k(M,\varphi) \to \Omega^{k+1}(M,\varphi),

        where `\Omega^k(M,\varphi)` denotes the space of differential forms of
        degree `k` along `\varphi`
        (see :meth:`~sage.manifolds.differentiable.diff_form.DiffForm.exterior_derivative`
        for further information). By linear extension, this induces a map on
        `\Omega^*(M,\varphi)`:

        .. MATH::

            \mathrm{d}: \Omega^*(M,\varphi) \to \Omega^*(M,\varphi).

        OUTPUT:

        - a :class:`MixedForm` representing the exterior
          derivative of the mixed form

        EXAMPLES:

        Exterior derivative of a mixed form on a 3-dimensional manifold::

            sage: M = Manifold(3, 'M', start_index=1)
            sage: c_xyz.<x,y,z> = M.chart()
            sage: f = M.scalar_field(z^2, name='f')
            sage: f.disp()
            f: M → ℝ
                (x, y, z) ↦ z^2
            sage: a = M.diff_form(2, 'a')
            sage: a[1,2], a[1,3], a[2,3] = z+y^2, z+x, x^2
            sage: a.disp()
            a = (y^2 + z) dx∧dy + (x + z) dx∧dz + x^2 dy∧dz
            sage: F = M.mixed_form([f, 0, a, 0], name='F'); F.display()
            F = f + zero + a + zero
            sage: dF = F.exterior_derivative()
            sage: dF.display()
            dF = zero + df + dzero + da
            sage: dF = F.exterior_derivative()
            sage: dF.display_expansion()
            dF = 2*z dz + (2*x + 1) dx∧dy∧dz

        Due to long calculation times, the result is cached::

            sage: F.exterior_derivative() is dF
            True
        """
        resu = self._new_instance()
        resu[0] = self._domain.zero_scalar_field()
        resu[1:] = [self[j].exterior_derivative()
                    for j in range(self._max_deg)]
        # Compose name:
        from sage.tensor.modules.format_utilities import (
            format_unop_latex,
            format_unop_txt,
        )
        resu._name = format_unop_txt('d', self._name)
        resu._latex_name = format_unop_latex(r'\mathrm{d}', self._latex_name)
        return resu

    derivative = exterior_derivative

    def copy(self, name=None, latex_name=None):
        r"""
        Return an exact copy of ``self``.

        .. NOTE::

            The name and names of the components are not copied.

        INPUT:

        - ``name`` -- (default: ``None``) name given to the copy
        - ``latex_name`` -- (default: ``None``) LaTeX symbol to denote the
          copy; if none is provided, the LaTeX symbol is set to ``name``

        EXAMPLES:

        Initialize a 2-dimensional manifold and differential forms::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U') ; V = M.open_subset('V')
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
            sage: xy_to_uv = c_xy.transition_map(c_uv, (x+y, x-y),
            ....:                    intersection_name='W', restrictions1= x>0,
            ....:                    restrictions2= u+v>0)
            sage: uv_to_xy = xy_to_uv.inverse()
            sage: W = U.intersection(V)
            sage: e_xy = c_xy.frame(); e_uv = c_uv.frame()
            sage: f = M.scalar_field(x, name='f', chart=c_xy)
            sage: f.add_expr_by_continuation(c_uv, W)
            sage: f.display()
            f: M → ℝ
            on U: (x, y) ↦ x
            on V: (u, v) ↦ 1/2*u + 1/2*v
            sage: omega = M.diff_form(1, name='omega')
            sage: omega[e_xy,0] = x
            sage: omega.add_comp_by_continuation(e_uv, W, c_uv)
            sage: omega.display()
            omega = x dx
            sage: A = M.mixed_form([f, omega, 0], name='A'); A.display()
            A = f + omega + zero
            sage: A.display_expansion(e_uv)
            A = 1/2*u + 1/2*v + (1/4*u + 1/4*v) du + (1/4*u + 1/4*v) dv

        An exact copy is made. The copy is an entirely new instance and has a
        different name, but has the very same values::

            sage: B = A.copy(); B.display()
            (unnamed scalar field) + (unnamed 1-form) + (unnamed 2-form)
            sage: B.display_expansion(e_uv)
            1/2*u + 1/2*v + (1/4*u + 1/4*v) du + (1/4*u + 1/4*v) dv
            sage: A == B
            True
            sage: A is B
            False
        """
        resu = self._new_instance()
        resu._comp = [form.copy() for form in self]
        resu.set_name(name=name, latex_name=latex_name)
        resu._is_zero = self._is_zero  # a priori

        return resu

    def __setitem__(self, index, values):
        r"""
        Set a component with respect to some vector frame.

        - ``index`` -- list of indices; if ``[:]`` is provided, all the
          components are set
        - ``values`` -- the values to be set

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: c_xy.<x,y> = M.chart()
            sage: f = M.scalar_field(x, name='f')
            sage: a = M.diff_form(1, name='a')
            sage: a[0] = y
            sage: b = M.diff_form(2, name='b')
            sage: b[0,1] = x*y
            sage: A = M.mixed_form([f, 0, 0], name='A'); A.display()
            A = f + zero + zero
            sage: A[1:3] = [a, b]; A.display()
            A = f + a + b
            sage: A.display_expansion()
            A = x + y dx + x*y dx∧dy
        """
        if self.is_immutable():
            raise ValueError("the components of an immutable element "
                             "cannot be changed")
        if isinstance(index, (int, Integer)):
            start, stop, step = index, index + 1, 1
        elif isinstance(index, slice):
            start, stop, step = index.indices(self._max_deg + 1)
        else:
            raise TypeError("index must be int, Integer or slice")
        if isinstance(values, list):
            form_list = values
        else:
            form_list = [values]
        if len(form_list) != len(range(start, stop, step)):
            raise IndexError("either input or index out of range")
        for deg, j in zip(range(start, stop, step), range(len(form_list))):
            dmodule = self._domain.diff_form_module(deg, self._dest_map)
            form = dmodule(form_list[j])
            self[deg].copy_from(form)  # keep the names
            self[deg].set_name(name=form._name, latex_name=form._latex_name)
        self._is_zero = False  # a priori

    def __getitem__(self, deg):
        r"""
        Return a component with respect to some frame.

        INPUT:

        - ``deg`` -- slice of which degrees shall be returned

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: c_xy.<x,y> = M.chart()
            sage: f = M.scalar_field(name='f')
            sage: a = M.diff_form(1, name='a')
            sage: b = M.diff_form(2, name='b')
            sage: A = M.mixed_form([f, a, b], name='A'); A.display()
            A = f + a + b
            sage: A.__getitem__(0)
            Scalar field f on the 2-dimensional differentiable manifold M
            sage: A.__getitem__(1)
            1-form a on the 2-dimensional differentiable manifold M
            sage: A.__getitem__(2)
            2-form b on the 2-dimensional differentiable manifold M
            sage: A.__getitem__(slice(0,3,1))
            [Scalar field f on the 2-dimensional differentiable manifold M,
             1-form a on the 2-dimensional differentiable manifold M,
             2-form b on the 2-dimensional differentiable manifold M]
        """
        if self._comp is None:
            self._init_comp()
        return self._comp[deg]

    def set_restriction(self, rst):
        r"""
        Set a (component-wise) restriction of ``self`` to some subdomain.

        INPUT:

        - ``rst`` -- :class:`MixedForm` of the same type as ``self``, defined on
          a subdomain of the domain of ``self``

        EXAMPLES:

        Initialize the 2-sphere::

            sage: M = Manifold(2, 'M') # the 2-dimensional sphere S^2
            sage: U = M.open_subset('U') # complement of the North pole
            sage: c_xy.<x,y> = U.chart() # stereographic coordinates from the North pole
            sage: V = M.open_subset('V') # complement of the South pole
            sage: c_uv.<u,v> = V.chart() # stereographic coordinates from the South pole
            sage: M.declare_union(U,V)   # S^2 is the union of U and V
            sage: xy_to_uv = c_xy.transition_map(c_uv, (x/(x^2+y^2), y/(x^2+y^2)),
            ....:                intersection_name='W', restrictions1= x^2+y^2!=0,
            ....:                restrictions2= u^2+v^2!=0)
            sage: uv_to_xy = xy_to_uv.inverse()
            sage: W = U.intersection(V)
            sage: e_xy = c_xy.frame(); e_uv = c_uv.frame()

        And define some forms on the subset ``U``::

            sage: f = U.scalar_field(x, name='f', chart=c_xy)
            sage: omega = U.diff_form(1, name='omega')
            sage: omega[e_xy,0] = y
            sage: AU = U.mixed_form([f, omega, 0], name='A'); AU
            Mixed differential form A on the Open subset U of the 2-dimensional
             differentiable manifold M
            sage: AU.display_expansion(e_xy)
            A = x + y dx

        A mixed form on ``M`` can be specified by some mixed form on a subset::

            sage: A = M.mixed_form(name='A'); A
            Mixed differential form A on the 2-dimensional differentiable
             manifold M
            sage: A.set_restriction(AU)
            sage: A.display_expansion(e_xy)
            A = x + y dx
            sage: A.add_comp_by_continuation(e_uv, W, c_uv)
            sage: A.display_expansion(e_uv)
            A = u/(u^2 + v^2) - (u^2*v - v^3)/(u^6 + 3*u^4*v^2 + 3*u^2*v^4 + v^6) du - 2*u*v^2/(u^6 + 3*u^4*v^2 + 3*u^2*v^4 + v^6) dv
            sage: A.restrict(U) == AU
            True
        """
        if not isinstance(rst, MixedForm):
            raise TypeError("the argument must be a mixed form")
        if not rst._domain.is_subset(self._domain):
            raise ValueError("the specified domain is not a subset of "
                             "the domain of definition of the mixed form")
        for j in self.irange():
            self[j].set_restriction(rst[j])
        self._is_zero = False  # a priori

    def restrict(self, subdomain, dest_map=None):
        r"""
        Return the restriction of ``self`` to some subdomain.

        INPUT:

        - ``subdomain`` --
          :class:`~sage.manifolds.differentiable.manifold.DifferentiableManifold`;
          open subset `U` of the domain of ``self``
        - ``dest_map`` --
          :class:`~sage.manifolds.differentiable.diff_map.DiffMap`
          (default: ``None``); destination map `\Psi:\ U \rightarrow V`,
          where `V` is an open subset of the manifold `N` where the mixed form
          takes it values; if ``None``, the restriction of `\Phi` to `U` is
          used, `\Phi` being the differentiable map `S \rightarrow M` associated
          with the mixed form

        OUTPUT: :class:`MixedForm` representing the restriction

        EXAMPLES:

        Initialize the 2-sphere::

            sage: M = Manifold(2, 'M') # the 2-dimensional sphere S^2
            sage: U = M.open_subset('U') # complement of the North pole
            sage: c_xy.<x,y> = U.chart() # stereographic coordinates from the North pole
            sage: V = M.open_subset('V') # complement of the South pole
            sage: c_uv.<u,v> = V.chart() # stereographic coordinates from the South pole
            sage: M.declare_union(U,V)   # S^2 is the union of U and V
            sage: xy_to_uv = c_xy.transition_map(c_uv, (x/(x^2+y^2), y/(x^2+y^2)),
            ....:                intersection_name='W', restrictions1= x^2+y^2!=0,
            ....:                restrictions2= u^2+v^2!=0)
            sage: uv_to_xy = xy_to_uv.inverse()
            sage: W = U.intersection(V)
            sage: e_xy = c_xy.frame(); e_uv = c_uv.frame()

        And predefine some forms::

            sage: f = M.scalar_field(x^2, name='f', chart=c_xy)
            sage: f.add_expr_by_continuation(c_uv, W)
            sage: omega = M.diff_form(1, name='omega')
            sage: omega[e_xy,0] = y^2
            sage: omega.add_comp_by_continuation(e_uv, W, c_uv)
            sage: eta = M.diff_form(2, name='eta')
            sage: eta[e_xy,0,1] = x^2*y^2
            sage: eta.add_comp_by_continuation(e_uv, W, c_uv)

        Now, a mixed form can be restricted to some subdomain::

            sage: F = M.mixed_form([f, omega, eta], name='F')
            sage: FV = F.restrict(V); FV
            Mixed differential form F on the Open subset V of the 2-dimensional
             differentiable manifold M
            sage: FV[:]
            [Scalar field f on the Open subset V of the 2-dimensional
             differentiable manifold M,
             1-form omega on the Open subset V of the 2-dimensional
             differentiable manifold M,
             2-form eta on the Open subset V of the 2-dimensional
             differentiable manifold M]
            sage: FV.display_expansion(e_uv)
            F = u^2/(u^4 + 2*u^2*v^2 + v^4) - (u^2*v^2 - v^4)/(u^8 + 4*u^6*v^2 + 6*u^4*v^4 + 4*u^2*v^6 + v^8) du - 2*u*v^3/(u^8 + 4*u^6*v^2 + 6*u^4*v^4 + 4*u^2*v^6 + v^8) dv - u^2*v^2/(u^12 + 6*u^10*v^2 + 15*u^8*v^4 + 20*u^6*v^6 + 15*u^4*v^8 + 6*u^2*v^10 + v^12) du∧dv
        """
        resu = type(self)(subdomain.mixed_form_algebra(dest_map=dest_map),
                          name=self._name, latex_name=self._latex_name)
        resu[0] = self[0].restrict(subdomain)
        resu[1:] = [self[j].restrict(subdomain, dest_map)
                    for j in self.irange(1)]
        return resu

    def add_comp_by_continuation(self, frame, subdomain, chart=None):
        r"""
        Set components with respect to a vector frame by continuation of the
        coordinate expression of the components in a subframe.

        The continuation is performed by demanding that the components have
        the same coordinate expression as those on the restriction of the
        frame to a given subdomain.

        INPUT:

        - ``frame`` -- vector frame `e` in which the components are to be set
        - ``subdomain`` -- open subset of `e`'s domain in which the
          components are known or can be evaluated from other components
        - ``chart`` -- (default: ``None``) coordinate chart on `e`'s domain in
          which the extension of the expression of the components is to be
          performed; if ``None``, the default's chart of `e`'s domain is
          assumed

        EXAMPLES:

        Mixed form defined by differential forms with components on different
        parts of the 2-sphere::

            sage: M = Manifold(2, 'M') # the 2-dimensional sphere S^2
            sage: U = M.open_subset('U') # complement of the North pole
            sage: c_xy.<x,y> = U.chart() # stereographic coordinates from the North pole
            sage: V = M.open_subset('V') # complement of the South pole
            sage: c_uv.<u,v> = V.chart() # stereographic coordinates from the South pole
            sage: M.declare_union(U,V)   # S^2 is the union of U and V
            sage: xy_to_uv = c_xy.transition_map(c_uv, (x/(x^2+y^2), y/(x^2+y^2)),
            ....:                intersection_name='W', restrictions1= x^2+y^2!=0,
            ....:                restrictions2= u^2+v^2!=0)
            sage: uv_to_xy = xy_to_uv.inverse()
            sage: W = U.intersection(V)
            sage: e_xy = c_xy.frame(); e_uv = c_uv.frame()
            sage: F = M.mixed_form(name='F') # No predefined components, here
            sage: F[0] = M.scalar_field(x, name='f')
            sage: F[1] = M.diff_form(1, {e_xy: [x,0]}, name='omega')
            sage: F[2].set_name(name='eta')
            sage: F[2][e_uv,0,1] = u*v
            sage: F.add_comp_by_continuation(e_uv, W, c_uv)
            sage: F.add_comp_by_continuation(e_xy, W, c_xy) # Now, F is fully defined
            sage: F.display_expansion(e_xy)
            F = x + x dx - x*y/(x^8 + 4*x^6*y^2 + 6*x^4*y^4 + 4*x^2*y^6 + y^8) dx∧dy
            sage: F.display_expansion(e_uv)
            F = u/(u^2 + v^2) - (u^3 - u*v^2)/(u^6 + 3*u^4*v^2 + 3*u^2*v^4 + v^6) du - 2*u^2*v/(u^6 + 3*u^4*v^2 + 3*u^2*v^4 + v^6) dv + u*v du∧dv
        """
        if chart is None:
            chart = frame._chart
        self[0].add_expr_by_continuation(chart, subdomain)
        for j in self.irange(1):
            self[j].add_comp_by_continuation(frame, subdomain, chart)

    def irange(self, start=None):
        r"""
        Single index generator.

        INPUT:

        - ``start`` -- (default: ``None``) initial value `i_0` of the index
          between 0 and `n`, where `n` is the manifold's dimension; if none is
          provided, the value 0 is assumed

        OUTPUT:

        - an iterable index, starting from `i_0` and ending at
          `n`, where `n` is the manifold's dimension

        EXAMPLES::

            sage: M = Manifold(3, 'M')
            sage: a = M.mixed_form(name='a')
            sage: list(a.irange())
            [0, 1, 2, 3]
            sage: list(a.irange(2))
            [2, 3]
        """
        return self.parent().irange(start=start)

    def set_comp(self, i):
        r"""
        Return the `i`-th homogeneous component for assignment.

        EXAMPLES::

            sage: M = Manifold(2, 'M')
            sage: X.<x,y> = M.chart()
            sage: A = M.mixed_form(name='A')
            sage: A.set_comp(0).set_expr(x^2) # scalar field
            sage: A.set_comp(1)[:] = [-y, x]
            sage: A.set_comp(2)[0,1] = x-y
            sage: A.display()
            A = A_0 + A_1 + A_2
            sage: A.display_expansion()
            A = x^2 - y dx + x dy + (x - y) dx∧dy
        """
        return self[i]

    def set_immutable(self):
        r"""
        Set ``self`` and homogeneous components of ``self`` immutable.

        EXAMPLES::

            sage: M = Manifold(2, 'M')
            sage: X.<x,y> = M.chart()
            sage: f = M.scalar_field(x^2, name='f')
            sage: A = M.mixed_form([f, 0, 0], name='A')
            sage: A.set_immutable()
            sage: A.is_immutable()
            True
            sage: A[0].is_immutable()
            True
            sage: f.is_immutable()
            False
        """
        for form in self:
            form.set_immutable()
        super().set_immutable()



# sage.doctest: needs sage.symbolic
"""
Plotting utilities
"""

# ****************************************************************************
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.ext.fast_callable import FastCallableFloatWrapper
from collections.abc import Iterable


def setup_for_eval_on_grid(funcs,
                           ranges,
                           plot_points=None,
                           return_vars=False,
                           imaginary_tolerance=1e-8):
    r"""
    Calculate the necessary parameters to construct a list of points,
    and make the functions fast_callable.

    INPUT:

    - ``funcs`` -- a function, or a list, tuple, or vector of functions

    - ``ranges`` -- list of ranges.  A range can be a 2-tuple of
      numbers specifying the minimum and maximum, or a 3-tuple giving
      the variable explicitly.

    - ``plot_points`` -- tuple of integers specifying the number of
      plot points for each range.  If a single number is specified, it
      will be the value for all ranges.  This defaults to 2.

    - ``return_vars`` -- (default: ``False``) if ``True``, return the variables,
      in order

    - ``imaginary_tolerance`` -- (default: ``1e-8``) if an imaginary
      number arises (due, for example, to numerical issues), this
      tolerance specifies how large it has to be in magnitude before
      we raise an error. In other words, imaginary parts smaller than
      this are ignored in your plot points.

    OUTPUT:

    - ``fast_funcs`` -- if only one function passed, then a fast
      callable function.  If funcs is a list or tuple, then a tuple
      of fast callable functions is returned.

    - ``range_specs`` -- list of range_specs: for each range, a
      tuple is returned of the form (range_min, range_max,
      range_step) such that ``srange(range_min, range_max,
      range_step, include_endpoint=True)`` gives the correct points
      for evaluation.

    EXAMPLES::

        sage: x,y,z = var('x,y,z')
        sage: f(x,y) = x+y-z
        sage: g(x,y) = x+y
        sage: h(y) = -y
        sage: sage.plot.misc.setup_for_eval_on_grid(f, [(0, 2),(1,3),(-4,1)], plot_points=5)
        (<sage...>, [(0.0, 2.0, 0.5), (1.0, 3.0, 0.5), (-4.0, 1.0, 1.25)])
        sage: sage.plot.misc.setup_for_eval_on_grid([g,h], [(0, 2),(-1,1)], plot_points=5)
        ((<sage...>, <sage...>), [(0.0, 2.0, 0.5), (-1.0, 1.0, 0.5)])
        sage: sage.plot.misc.setup_for_eval_on_grid([sin,cos], [(-1,1)], plot_points=9)
        ((<sage...>, <sage...>), [(-1.0, 1.0, 0.25)])
        sage: sage.plot.misc.setup_for_eval_on_grid([lambda x: x^2,cos], [(-1,1)], plot_points=9)
        ((<function <lambda> ...>, <sage...>), [(-1.0, 1.0, 0.25)])
        sage: sage.plot.misc.setup_for_eval_on_grid([x+y], [(x,-1,1),(y,-2,2)])
        ((<sage...>,), [(-1.0, 1.0, 2.0), (-2.0, 2.0, 4.0)])
        sage: sage.plot.misc.setup_for_eval_on_grid(x+y, [(x,-1,1),(y,-1,1)], plot_points=[4,9])
        (<sage...>, [(-1.0, 1.0, 0.6666666666666666), (-1.0, 1.0, 0.25)])
        sage: sage.plot.misc.setup_for_eval_on_grid(x+y, [(x,-1,1),(y,-1,1)], plot_points=[4,9,10])
        Traceback (most recent call last):
        ...
        ValueError: plot_points must be either an integer or a list of integers, one for each range
        sage: sage.plot.misc.setup_for_eval_on_grid(x+y, [(1,-1),(y,-1,1)], plot_points=[4,9,10])
        Traceback (most recent call last):
        ...
        ValueError: Some variable ranges specify variables while others do not

    Beware typos: a comma which should be a period, for instance::

        sage: sage.plot.misc.setup_for_eval_on_grid(x+y, [(x, 1, 2), (y, 0,1, 0.2)], plot_points=[4,9,10])
        Traceback (most recent call last):
        ...
        ValueError: At least one variable range has more than 3 entries: each should either have 2 or 3 entries, with one of the forms (xmin, xmax) or (x, xmin, xmax)

        sage: sage.plot.misc.setup_for_eval_on_grid(x+y, [(y,1,-1),(x,-1,1)], plot_points=5)
        (<sage...>, [(-1.0, 1.0, 0.5), (-1.0, 1.0, 0.5)])
        sage: sage.plot.misc.setup_for_eval_on_grid(x+y, [(x,1,-1),(x,-1,1)], plot_points=5)
        Traceback (most recent call last):
        ...
        ValueError: range variables should be distinct, but there are duplicates
        sage: sage.plot.misc.setup_for_eval_on_grid(x+y, [(x,1,1),(y,-1,1)])
        Traceback (most recent call last):
        ...
        ValueError: plot start point and end point must be different
        sage: sage.plot.misc.setup_for_eval_on_grid(x+y, [(x,1,-1),(y,-1,1)], return_vars=True)
        (<sage...>, [(-1.0, 1.0, 2.0), (-1.0, 1.0, 2.0)], [x, y])
        sage: sage.plot.misc.setup_for_eval_on_grid(x+y, [(y,1,-1),(x,-1,1)], return_vars=True)
        (<sage...>, [(-1.0, 1.0, 2.0), (-1.0, 1.0, 2.0)], [y, x])

    TESTS:

    Ensure that we can plot expressions with intermediate complex
    terms as in :issue:`8450`::

        sage: x, y = SR.var('x y')
        sage: contour_plot(abs(x+i*y), (x,-1,1), (y,-1,1))
        Graphics object consisting of 1 graphics primitive
        sage: density_plot(abs(x+i*y), (x,-1,1), (y,-1,1))
        Graphics object consisting of 1 graphics primitive
        sage: plot3d(abs(x+i*y), (x,-1,1),(y,-1,1))
        Graphics3d Object
        sage: streamline_plot(abs(x+i*y), (x,-1,1),(y,-1,1))
        Graphics object consisting of 1 graphics primitive
    """
    if max(map(len, ranges)) > 3:
        raise ValueError("At least one variable range has more than 3 entries: each should either have 2 or 3 entries, with one of the forms (xmin, xmax) or (x, xmin, xmax)")
    if max(map(len, ranges)) != min(map(len, ranges)):
        raise ValueError("Some variable ranges specify variables while others do not")

    if len(ranges[0]) == 3:
        vars = [r[0] for r in ranges]
        ranges = [r[1:] for r in ranges]
        if len(set(vars)) < len(vars):
            raise ValueError("range variables should be distinct, but there are duplicates")
    else:
        vars, free_vars = unify_arguments(funcs)

    # check for invalid range (xmin > xmax or ymin > ymax) and swap
    if len(ranges) > 1:
        for i in range(len(ranges)):
            if ranges[i][-2] > ranges[i][-1]:
                ranges[i] = list(ranges[i])
                ranges[i][-1], ranges[i][-2] = ranges[i][-2], ranges[i][-1]
                ranges[i] = tuple(ranges[i])

    # pad the variables if we don't have enough
    nargs = len(ranges)
    if len(vars) < nargs:
        vars += ('_',) * (nargs - len(vars))

    ranges = [[float(z) for z in r] for r in ranges]

    if plot_points is None:
        plot_points = 2

    if not isinstance(plot_points, (list, tuple)):
        plot_points = [plot_points] * len(ranges)
    elif len(plot_points) != nargs:
        raise ValueError("plot_points must be either an integer or a list of integers, one for each range")

    plot_points = [int(p) if p >= 2 else 2 for p in plot_points]
    range_steps = [abs(range[1] - range[0]) / (p - 1)
                   for range, p in zip(ranges, plot_points)]
    if min(range_steps) == float(0):
        raise ValueError("plot start point and end point must be different")

    eov = False  # eov = "expect one value"
    if nargs == 1:
        eov = True

    from sage.ext.fast_callable import fast_callable

    def try_make_fast(f):
        # If "f" supports fast_callable(), use it. We can't guarantee
        # that our arguments will actually support fast_callable()
        # because, for example, the user may already have done it
        # himself, and the result of fast_callable() can't be
        # fast-callabled again.
        from sage.rings.complex_double import CDF
        from sage.ext.interpreters.wrapper_cdf import Wrapper_cdf

        if hasattr(f, '_fast_callable_'):
            ff = fast_callable(f, vars=vars, expect_one_var=eov, domain=CDF)
            return FastCallablePlotWrapper(ff, imag_tol=imaginary_tolerance)
        if isinstance(f, Wrapper_cdf):
            # Already a fast-callable, just wrap it. This can happen
            # if, for example, a symbolic expression is passed to a
            # higher-level plot() function that converts it to a
            # fast-callable with expr._plot_fast_callable() before
            # we ever see it.
            return FastCallablePlotWrapper(f, imag_tol=imaginary_tolerance)
        if callable(f):
            # This will catch python functions, among other things. We don't
            # wrap these yet because we don't know what type they'll return.
            return f
        # Convert things like ZZ(0) into constant functions.
        from sage.symbolic.ring import SR
        ff = fast_callable(SR(f),
                           vars=vars,
                           expect_one_var=eov,
                           domain=CDF)
        return FastCallablePlotWrapper(ff, imag_tol=imaginary_tolerance)

    # Handle vectors, lists, tuples, etc.
    if isinstance(funcs, Iterable):
        funcs = tuple(try_make_fast(f) for f in funcs)
    else:
        funcs = try_make_fast(funcs)

    # TODO: raise an error if there is a function/method in funcs that
    # takes more values than we have ranges

    if return_vars:
        return (funcs,
                [tuple(_range + [range_step])
                 for _range, range_step in zip(ranges, range_steps)],
                vars)
    return (funcs,
            [tuple(_range + [range_step])
             for _range, range_step in zip(ranges, range_steps)])


def unify_arguments(funcs):
    """
    Return a tuple of variables of the functions, as well as the
    number of "free" variables (i.e., variables that defined in a
    callable function).

    INPUT:

    - ``funcs`` -- list of functions; these can be symbolic
      expressions, polynomials, etc.

    OUTPUT: functions, expected arguments

    - A tuple of variables in the functions

    - A tuple of variables that were "free" in the functions

    EXAMPLES::

        sage: x,y,z = var('x,y,z')
        sage: f(x,y) = x+y-z
        sage: g(x,y) = x+y
        sage: h(y) = -y
        sage: sage.plot.misc.unify_arguments((f,g,h))
        ((x, y, z), (z,))
        sage: sage.plot.misc.unify_arguments((g,h))
        ((x, y), ())
        sage: sage.plot.misc.unify_arguments((f,z))
        ((x, y, z), (z,))
        sage: sage.plot.misc.unify_arguments((h,z))
        ((y, z), (z,))
        sage: sage.plot.misc.unify_arguments((x+y,x-y))
        ((x, y), (x, y))
    """
    vars = set()
    free_variables = set()
    if not isinstance(funcs, (list, tuple)):
        funcs = [funcs]

    from sage.structure.element import Expression
    for f in funcs:
        if isinstance(f, Expression) and f.is_callable():
            f_args = set(f.arguments())
            vars.update(f_args)
        else:
            f_args = set()

        try:
            free_vars = set(f.variables()).difference(f_args)
            vars.update(free_vars)
            free_variables.update(free_vars)
        except AttributeError:
            # we probably have a constant
            pass
    return tuple(sorted(vars, key=str)), tuple(sorted(free_variables, key=str))


def _multiple_of_constant(n, pos, const):
    r"""
    Function for internal use in formatting ticks on axes with
    nice-looking multiples of various symbolic constants, such
    as `\pi` or `e`.

    This should only be used via keyword argument
    ``tick_formatter`` in :meth:`plot.show`.  See documentation
    for the matplotlib.ticker module for more details.

    EXAMPLES:

    Here is the intended use::

        sage: plot(sin(x), (x,0,2*pi), ticks=pi/3, tick_formatter=pi)
        Graphics object consisting of 1 graphics primitive

    Here is an unintended use, which yields unexpected (and probably
    undesired) results::

        sage: plot(x^2, (x, -2, 2), tick_formatter=pi)
        Graphics object consisting of 1 graphics primitive

    We can also use more unusual constant choices::

        sage: plot(ln(x), (x,0,10), ticks=e, tick_formatter=e)
        Graphics object consisting of 1 graphics primitive
        sage: plot(x^2, (x,0,10), ticks=[sqrt(2),8], tick_formatter=sqrt(2))
        Graphics object consisting of 1 graphics primitive
    """
    from sage.misc.latex import latex
    from sage.rings.continued_fraction import continued_fraction
    from sage.rings.infinity import Infinity
    cf = continued_fraction(n / const)
    k = 1
    while cf.quotient(k) != Infinity and cf.denominator(k) < 12:
        k += 1
    return '$%s$' % latex(cf.convergent(k - 1) * const)


def get_matplotlib_linestyle(linestyle, return_type):
    """
    Function which translates between matplotlib linestyle in short notation
    (i.e. '-', '--', ':', '-.') and long notation (i.e. 'solid', 'dashed',
    'dotted', 'dashdot' ).

    If linestyle is none of these allowed options, the function raises
    a :exc:`ValueError`.

    INPUT:

    - ``linestyle`` -- the style of the line, which is one of
       - ``'-'`` or ``'solid'``
       - ``'--'`` or ``'dashed'``
       - ``'-.'`` or ``'dash dot'``
       - ``':'`` or ``'dotted'``
       - ``"None"`` or ``" "`` or ``""`` (nothing)

       The linestyle can also be prefixed with a drawing style (e.g., ``'steps--'``)

       - ``'default'`` (connect the points with straight lines)
       - ``'steps'`` or ``'steps-pre'`` (step function; horizontal
         line is to the left of point)
       - ``'steps-mid'`` (step function; points are in the middle of
         horizontal lines)
       - ``'steps-post'`` (step function; horizontal line is to the
         right of point)

       If ``linestyle`` is ``None`` (of type NoneType), then we return it
       back unmodified.

    - ``return_type`` -- the type of linestyle that should be output. This
      argument takes only two values - ``'long'`` or ``'short'``

    EXAMPLES:

    Here is an example how to call this function::

        sage: from sage.plot.misc import get_matplotlib_linestyle
        sage: get_matplotlib_linestyle(':', return_type='short')
        ':'

        sage: get_matplotlib_linestyle(':', return_type='long')
        'dotted'

    TESTS:

    Make sure that if the input is already in the desired format, then it
    is unchanged::

        sage: get_matplotlib_linestyle(':', 'short')
        ':'

    Empty linestyles should be handled properly::

        sage: get_matplotlib_linestyle("", 'short')
        ''
        sage: get_matplotlib_linestyle("", 'long')
        'None'
        sage: get_matplotlib_linestyle(None, 'short') is None
        True

    Linestyles with ``'default'`` or ``'steps'`` in them should also be
    properly handled.  For instance, matplotlib understands only the short
    version when ``'steps'`` is used::

        sage: get_matplotlib_linestyle("default", "short")
        ''
        sage: get_matplotlib_linestyle("steps--", "short")
        'steps--'
        sage: get_matplotlib_linestyle("steps-predashed", "long")
        'steps-pre--'

    Finally, raise error on invalid linestyles::

        sage: get_matplotlib_linestyle("isthissage", "long")
        Traceback (most recent call last):
        ...
        ValueError: WARNING: Unrecognized linestyle 'isthissage'. Possible
        linestyle options are:
        {'solid', 'dashed', 'dotted', dashdot', 'None'}, respectively {'-',
        '--', ':', '-.', ''}
    """
    long_to_short_dict = {'solid': '-',
                          'dashed': '--',
                          'dotted': ':',
                          'dashdot': '-.'}
    short_to_long_dict = {'-': 'solid',
                          '--': 'dashed',
                          ':': 'dotted',
                          '-.': 'dashdot'}

    # We need this to take care of region plot. Essentially, if None is
    # passed, then we just return back the same thing.
    if linestyle is None:
        return None

    if linestyle.startswith("default"):
        return get_matplotlib_linestyle(linestyle.removeprefix("default"), "short")
    if linestyle.startswith("steps"):
        if linestyle.startswith("steps-mid"):
            return "steps-mid" + get_matplotlib_linestyle(
                linestyle.removeprefix("steps-mid"), "short")
        if linestyle.startswith("steps-post"):
            return "steps-post" + get_matplotlib_linestyle(
                linestyle.removeprefix("steps-post"), "short")
        if linestyle.startswith("steps-pre"):
            return "steps-pre" + get_matplotlib_linestyle(
                linestyle.removeprefix("steps-pre"), "short")
        return "steps" + get_matplotlib_linestyle(
            linestyle.removeprefix("steps"), "short")

    if return_type == 'short':
        if linestyle in short_to_long_dict.keys():
            return linestyle
        if linestyle == "" or linestyle == " " or linestyle == "None":
            return ''
        if linestyle in long_to_short_dict.keys():
            return long_to_short_dict[linestyle]
        raise ValueError("WARNING: Unrecognized linestyle '%s'. "
                         "Possible linestyle options are:\n{'solid', "
                         "'dashed', 'dotted', dashdot', 'None'}, "
                         "respectively {'-', '--', ':', '-.', ''}" %
                         (linestyle))

    elif return_type == 'long':
        if linestyle in long_to_short_dict.keys():
            return linestyle
        if linestyle == "" or linestyle == " " or linestyle == "None":
            return "None"
        if linestyle in short_to_long_dict.keys():
            return short_to_long_dict[linestyle]
        raise ValueError("WARNING: Unrecognized linestyle '%s'. "
                         "Possible linestyle options are:\n{'solid', "
                         "'dashed', 'dotted', dashdot', 'None'}, "
                         "respectively {'-', '--', ':', '-.', ''}" %
                         (linestyle))


class FastCallablePlotWrapper(FastCallableFloatWrapper):
    r"""
    A fast-callable wrapper for plotting that returns ``nan`` instead
    of raising an error whenever the imaginary tolerance is exceeded.

    A detailed rationale for this can be found in the superclass
    documentation.

    EXAMPLES:

    The ``float`` incarnation of "not a number" is returned instead
    of an error being thrown if the answer is complex::

        sage: from sage.plot.misc import FastCallablePlotWrapper
        sage: f = sqrt(x)
        sage: ff = fast_callable(f, vars=[x], domain=CDF)
        sage: fff = FastCallablePlotWrapper(ff, imag_tol=1e-8)
        sage: fff(1)
        1.0
        sage: fff(-1)
        nan
    """
    def __call__(self, *args):
        r"""
        Evaluate the underlying fast-callable and convert the result to
        ``float``.

        TESTS:

        Evaluation never fails and always returns a ``float``::

            sage: from sage.plot.misc import FastCallablePlotWrapper
            sage: f = x
            sage: ff = fast_callable(f, vars=[x], domain=CDF)
            sage: fff = FastCallablePlotWrapper(ff, imag_tol=0.1)
            sage: type(fff(CDF.random_element())) is float
            True
        """
        try:
            return super().__call__(*args)
        except ValueError:
            return float("nan")
